import time
import warnings
from itertools import combinations_with_replacement, product
from queue import PriorityQueue
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from formulaic import Formula
from scipy.optimize._minimize import standardize_constraints

from bofire.data_models.constraints.api import (
    ConstraintNotFulfilledError,
    NChooseKConstraint,
    NonlinearConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.enum import SamplingMethodEnum
from bofire.data_models.features.api import ContinuousInput, Input
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.doe.objective import get_objective_class
from bofire.strategies.doe.utils import (
    constraints_as_scipy_constraints,
    get_formula_from_string,
    metrics,
    nchoosek_constraints_as_bounds,
)
from bofire.strategies.enum import OptimalityCriterionEnum
from bofire.strategies.samplers.polytope import PolytopeSampler


def find_local_max_ipopt_BaB(
    domain: Domain,
    model_type: Union[str, Formula],
    n_experiments: Optional[int] = None,
    delta: float = 1e-7,
    ipopt_options: Optional[Dict] = None,
    sampling: Optional[pd.DataFrame] = None,
    fixed_experiments: Optional[pd.DataFrame] = None,
    partially_fixed_experiments: Optional[pd.DataFrame] = None,
    objective: OptimalityCriterionEnum = OptimalityCriterionEnum.D_OPTIMALITY,
    categorical_groups: Optional[List[List[ContinuousInput]]] = None,
    discrete_variables: Optional[Dict[str, Tuple[ContinuousInput, List[float]]]] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    """Function computing a d-optimal design" for a given domain and model.
    It allows for the problem to have categorical values which is solved by Branch-and-Bound
        Args:
            domain (Domain): domain containing the inputs and constraints.
            model_type (str, Formula): keyword or formulaic Formula describing the model. Known keywords
                are "linear", "linear-and-interactions", "linear-and-quadratic", "fully-quadratic".
            n_experiments (int): Number of experiments. By default the value corresponds to
                the number of model terms - dimension of ker() + 3.
            delta (float): Regularization parameter. Default value is 1e-3.
            ipopt_options (Dict, optional): options for IPOPT. For more information see [this link](https://coin-or.github.io/Ipopt/OPTIONS.html)
            sampling (pd.DataFrame): dataframe containing the initial guess.
            fixed_experiments (pd.DataFrame): dataframe containing experiments that will be definitely part of the design.
                Values are set before the optimization.
            partially_fixed_experiments (pd.DataFrame): dataframe containing (some) fixed variables for experiments.
                Values are set before the optimization. Within one experiment not all variables need to be fixed.
                Variables can be fixed to one value or can be set to a range by setting a tuple with lower and upper bound
                Non-fixed variables have to be set to None or nan.
            objective (OptimalityCriterionEnum): OptimalityCriterionEnum object indicating which objective function to use.
            categorical_groups (Optional[List[List[ContinuousInput]]]). Represents the different groups of the
               relaxed categorical variables. Defaults to None.
            discrete_variables (Optional[Dict[str, Tuple[ContinuousInput, List[float]]]]): dict of relaxed discrete inputs
                with key:(relaxed variable, valid values). Defaults to None
            verbose (bool): if true, print information during the optimization process
        Returns:
            A pd.DataFrame object containing the best found input for the experiments. In general, this is only a
            local optimum.
    """
    from bofire.strategies.doe.branch_and_bound import NodeExperiment, bnb

    if categorical_groups is None:
        categorical_groups = []

    n_experiments = get_n_experiments(
        domain=domain, model_type=model_type, n_experiments=n_experiments
    )

    # get objective function
    model_formula = get_formula_from_string(
        model_type=model_type, rhs_only=True, domain=domain
    )
    objective_class = get_objective_class(objective)
    objective_class = objective_class(
        domain=domain, model=model_formula, n_experiments=n_experiments, delta=delta
    )

    # setting up initial node in the branch-and-bound tree
    column_keys = domain.inputs.get_keys()

    if fixed_experiments is not None:
        subtract = len(fixed_experiments)
        initial_branch = pd.DataFrame(
            np.full((n_experiments - subtract, len(column_keys)), None),
            columns=column_keys,
        )
        initial_branch = pd.concat([fixed_experiments, initial_branch]).reset_index(
            drop=True
        )
    else:
        initial_branch = pd.DataFrame(
            np.full((n_experiments, len(column_keys)), None),
            columns=column_keys,
        )

    if partially_fixed_experiments is not None:
        partially_fixed_experiments = pd.concat(
            [
                partially_fixed_experiments,
                pd.DataFrame(
                    np.full(
                        (
                            n_experiments - len(partially_fixed_experiments),
                            len(domain.inputs),
                        ),
                        None,
                    ),
                    columns=domain.get_feature_keys(includes=Input),
                ),
            ]
        ).reset_index(drop=True)

        initial_branch.mask(
            partially_fixed_experiments.notnull(),  # type: ignore
            other=partially_fixed_experiments,
            inplace=True,
        )

    initial_design = find_local_max_ipopt(
        domain,
        model_type,
        n_experiments,
        delta,
        ipopt_options,
        sampling,
        None,
        partially_fixed_experiments=initial_branch,
        objective=objective,
    )
    initial_value = objective_class.evaluate(
        initial_design.to_numpy().flatten(),
    )

    initial_node = NodeExperiment(
        initial_branch,
        initial_design,
        initial_value,
        categorical_groups,
        discrete_variables,  # type: ignore
    )

    # initializing branch-and-bound queue
    initial_queue = PriorityQueue()
    initial_queue.put(initial_node)

    # starting branch-and-bound
    result_node = bnb(
        initial_queue,
        domain=domain,
        model_type=model_type,
        n_experiments=n_experiments,
        delta=delta,
        ipopt_options=ipopt_options,
        sampling=sampling,
        fixed_experiments=None,
        objective=objective,
        verbose=verbose,
    )

    return result_node.design_matrix


def find_local_max_ipopt_exhaustive(
    domain: Domain,
    model_type: Union[str, Formula],
    n_experiments: Optional[int] = None,
    delta: float = 1e-7,
    ipopt_options: Optional[Dict] = None,
    sampling: Optional[pd.DataFrame] = None,
    fixed_experiments: Optional[pd.DataFrame] = None,
    objective: OptimalityCriterionEnum = OptimalityCriterionEnum.D_OPTIMALITY,
    partially_fixed_experiments: Optional[pd.DataFrame] = None,
    categorical_groups: Optional[List[List[ContinuousInput]]] = None,
    discrete_variables: Optional[Dict[str, Tuple[ContinuousInput, List[float]]]] = None,
    verbose: bool = False,
) -> pd.DataFrame:
    """Function computing a d-optimal design" for a given domain and model.
    It allows for the problem to have categorical values which is solved by exhaustive search
        Args:
            domain (Domain): domain containing the inputs and constraints.
            model_type (str, Formula): keyword or formulaic Formula describing the model. Known keywords
                are "linear", "linear-and-interactions", "linear-and-quadratic", "fully-quadratic".
            n_experiments (int): Number of experiments. By default the value corresponds to
                the number of model terms - dimension of ker() + 3.
            delta (float): Regularization parameter. Default value is 1e-3.
            ipopt_options (Dict, optional): options for IPOPT. For more information see [this link](https://coin-or.github.io/Ipopt/OPTIONS.html)
            sampling (pd.DataFrame): dataframe containing the initial guess.
            fixed_experiments (pd.DataFrame): dataframe containing experiments that will be definitely part of the design.
                Values are set before the optimization.
            objective (OptimalityCriterionEnum): OptimalityCriterionEnum object indicating which objective function to use.
            partially_fixed_experiments (pd.DataFrame): dataframe containing (some) fixed variables for experiments.
                Values are set before the optimization. Within one experiment not all variables need to be fixed.
                Variables can be fixed to one value or can be set to a range by setting a tuple with lower and upper bound
                Non-fixed variables have to be set to None or nan.
            categorical_groups (Optional[List[List[ContinuousInput]]]). Represents the different groups of the
               relaxed categorical variables. Defaults to None.
            discrete_variables (Optional[Dict[str, Tuple[ContinuousInput, List[float]]]]): dict of relaxed discrete inputs
                with key:(relaxed variable, valid values). Defaults to None
            verbose (bool): if true, print information during the optimization process
        Returns:
            A pd.DataFrame object containing the best found input for the experiments. In general, this is only a
            local optimum.
    """

    if categorical_groups is None:
        categorical_groups = []

    if discrete_variables is not None or len(discrete_variables) > 0:  # type: ignore
        raise NotImplementedError(
            "Exhaustive search for discrete variables is not implemented yet."
        )

    # get objective function
    model_formula = get_formula_from_string(
        model_type=model_type, rhs_only=True, domain=domain
    )
    objective_class = get_objective_class(objective)
    objective_class = objective_class(
        domain=domain, model=model_formula, n_experiments=n_experiments, delta=delta
    )

    # get binary variables
    binary_vars = [var for group in categorical_groups for var in group]
    list_keys = [var.key for var in binary_vars]

    # determine possible fixations of the different categories
    allowed_fixations = []
    for group in categorical_groups:
        allowed_fixations.append(np.eye(len(group)))

    n_experiments = get_n_experiments(
        domain=domain, model_type=model_type, n_experiments=n_experiments
    )
    n_non_fixed_experiments = n_experiments
    if fixed_experiments is not None:
        n_non_fixed_experiments -= len(fixed_experiments)

    allowed_fixations = product(*allowed_fixations)
    all_n_fixed_experiments = combinations_with_replacement(
        allowed_fixations, n_non_fixed_experiments
    )

    if partially_fixed_experiments is not None:
        partially_fixed_experiments = pd.concat(
            [
                partially_fixed_experiments,
                pd.DataFrame(
                    np.full(
                        (
                            n_non_fixed_experiments - len(partially_fixed_experiments),
                            len(domain.inputs),
                        ),
                        None,
                    ),
                    columns=domain.get_feature_keys(includes=Input),
                ),
            ]
        ).reset_index(drop=True)

    # testing all different fixations
    column_keys = domain.inputs.get_keys()
    group_keys = [var.key for group in categorical_groups for var in group]
    minimum = float("inf")
    optimal_design = pd.DataFrame()
    all_n_fixed_experiments = list(all_n_fixed_experiments)
    for i, binary_fixed_experiments in enumerate(all_n_fixed_experiments):
        if verbose:
            start_time = time.time()
        # setting up the pd.Dataframe for the partially fixed experiment
        binary_fixed_experiments = np.array(
            [
                var
                for experiment in binary_fixed_experiments
                for group in experiment
                for var in group
            ]
        ).reshape(n_non_fixed_experiments, len(binary_vars))

        binary_fixed_experiments = pd.DataFrame(
            binary_fixed_experiments, columns=group_keys
        )
        one_set_of_experiments = pd.DataFrame(
            np.full((n_non_fixed_experiments, len(domain.inputs)), None),
            columns=column_keys,
        )

        one_set_of_experiments.mask(
            binary_fixed_experiments.notnull(),
            other=binary_fixed_experiments,
            inplace=True,
        )

        if partially_fixed_experiments is not None:
            one_set_of_experiments.mask(
                partially_fixed_experiments.notnull(),
                other=partially_fixed_experiments,
                inplace=True,
            )

        if fixed_experiments is not None:
            one_set_of_experiments = pd.concat(
                [fixed_experiments, one_set_of_experiments]
            ).reset_index(drop=True)

        if sampling is not None:
            sampling.loc[:, list_keys] = one_set_of_experiments[list_keys].to_numpy()

        # minimizing with the current fixation
        try:
            current_design = find_local_max_ipopt(
                domain,
                model_type,
                n_experiments,
                delta,
                ipopt_options,
                sampling,
                None,
                one_set_of_experiments,
                objective,
            )
            domain.validate_candidates(
                candidates=current_design.apply(lambda x: np.round(x, 8)),
                only_inputs=True,
                tol=1e-4,
                raise_validation_error=True,
            )
            temp_value = objective_class.evaluate(
                current_design.to_numpy().flatten(),
            )
            if minimum is None or minimum > temp_value:
                minimum = temp_value
                optimal_design = current_design
            if verbose:
                print(
                    f"branch: {i} / {len(all_n_fixed_experiments)}, time: {time.time() - start_time} solution: {temp_value}, minimum after run {minimum}, difference: {temp_value - minimum}"  # type: ignore
                )
        except ConstraintNotFulfilledError:
            if verbose:
                print("skipping branch because of not fulfilling constraints")
    return optimal_design


def find_local_max_ipopt(
    domain: Domain,
    model_type: Union[str, Formula],
    n_experiments: Optional[int] = None,
    delta: float = 1e-7,
    ipopt_options: Optional[Dict] = None,
    sampling: Optional[pd.DataFrame] = None,
    fixed_experiments: Optional[pd.DataFrame] = None,
    partially_fixed_experiments: Optional[pd.DataFrame] = None,
    objective: OptimalityCriterionEnum = OptimalityCriterionEnum.D_OPTIMALITY,
) -> pd.DataFrame:
    """Function computing an optimal design for a given domain and model.
    Args:
        domain (Domain): domain containing the inputs and constraints.
        model_type (str, Formula): keyword or formulaic Formula describing the model. Known keywords
            are "linear", "linear-and-interactions", "linear-and-quadratic", "fully-quadratic".
        n_experiments (int): Number of experiments. By default the value corresponds to
            the number of model terms - dimension of ker() + 3.
        delta (float): Regularization parameter. Default value is 1e-3.
        ipopt_options (Dict, optional): options for IPOPT. For more information see [this link](https://coin-or.github.io/Ipopt/OPTIONS.html)
        sampling (pd.DataFrame): dataframe containing the initial guess.
        fixed_experiments (pd.DataFrame): dataframe containing experiments that will be definitely part of the design.
            Values are set before the optimization.
        partially_fixed_experiments (pd.DataFrame): dataframe containing (some) fixed variables for experiments.
            Values are set before the optimization. Within one experiment not all variables need to be fixed.
            Variables can be fixed to one value or can be set to a range by setting a tuple with lower and upper bound
            Non-fixed variables have to be set to None or nan.
        objective (OptimalityCriterionEnum): OptimalityCriterionEnum object indicating which objective function to use.
    Returns:
        A pd.DataFrame object containing the best found input for the experiments. In general, this is only a
        local optimum.
    """

    #
    # Checks and preparation steps
    #

    # warn user if IPOPT scipy interface is not available
    try:
        from cyipopt import minimize_ipopt  # type: ignore
    except ImportError as e:
        warnings.warn(e.msg)
        warnings.warn(
            "please run `conda install -c conda-forge cyipopt` for this functionality."
        )
        raise e

    # warn user about usage of nonlinear constraints
    if domain.constraints:
        if np.any([isinstance(c, NonlinearConstraint) for c in domain.constraints]):
            warnings.warn(
                "Nonlinear constraints were detected. Not all features and checks are supported for this type of constraints. \
                Using them can lead to unexpected behaviour. Please make sure to provide jacobians for nonlinear constraints.",
                UserWarning,
            )

    # check that NChooseK constraints only impose an upper bound on the number of nonzero components (and no lower bound)
    assert all(
        c.min_count == 0
        for c in domain.constraints
        if isinstance(c, NChooseKConstraint)
    ), "NChooseKConstraint with min_count !=0 is not supported!"

    # determine number of experiments (only relevant if n_experiments is not provided by the user)
    n_experiments = get_n_experiments(
        domain=domain, model_type=model_type, n_experiments=n_experiments
    )

    #
    # Sampling initital values
    #

    if sampling is not None:
        sampling.sort_index(axis=1, inplace=True)
        x0 = sampling.values.flatten()
    else:
        if len(domain.constraints.get(NonlinearConstraint)) == 0:
            sampler = PolytopeSampler(
                data_model=PolytopeSamplerDataModel(domain=domain)
            )
            x0 = sampler.ask(n_experiments, return_all=False).to_numpy().flatten()
        else:
            warnings.warn(
                "Sampling failed. Falling back to uniform sampling on input domain.\
                          Providing a custom sampling strategy compatible with the problem can \
                          possibly improve performance."
            )
            x0 = (
                domain.inputs.sample(n=n_experiments, method=SamplingMethodEnum.UNIFORM)
                .to_numpy()
                .flatten()
            )

    # get objective function and its jacobian
    model_formula = get_formula_from_string(
        model_type=model_type, rhs_only=True, domain=domain
    )

    objective_class = get_objective_class(objective)
    d_optimality = objective_class(
        domain=domain, model=model_formula, n_experiments=n_experiments, delta=delta
    )

    # write constraints as scipy constraints
    constraints = constraints_as_scipy_constraints(
        domain, n_experiments, ignore_nchoosek=True
    )

    # find bounds imposing NChooseK constraints
    bounds = nchoosek_constraints_as_bounds(domain, n_experiments)

    # fix experiments if any are given
    if fixed_experiments is not None:
        fixed_experiments.sort_index(axis=1, inplace=True)
        domain.validate_candidates(fixed_experiments, only_inputs=True)
        for i, val in enumerate(fixed_experiments.values.flatten()):
            bounds[i] = (val, val)
            x0[i] = val

    # partially fix experiments if any are given
    bounds, x0 = partially_fix_experiment(
        bounds, fixed_experiments, n_experiments, partially_fixed_experiments, x0
    )

    # set ipopt options
    if ipopt_options is None:
        ipopt_options = {}
    _ipopt_options = {"maxiter": 500, "disp": 0}
    for key in ipopt_options.keys():
        _ipopt_options[key] = ipopt_options[key]
    if _ipopt_options["disp"] > 12:
        _ipopt_options["disp"] = 0

    #
    # Do the optimization
    #

    result = minimize_ipopt(
        d_optimality.evaluate,
        x0=x0,
        bounds=bounds,
        # "SLSQP" has no deeper meaning here and just ensures correct constraint standardization
        constraints=standardize_constraints(constraints, x0, "SLSQP"),
        options=_ipopt_options,
        jac=d_optimality.evaluate_jacobian,
    )

    design = pd.DataFrame(
        result["x"].reshape(n_experiments, len(domain.inputs)),
        columns=domain.inputs.get_keys(),
        index=[f"exp{i}" for i in range(n_experiments)],
    )

    # exit message
    if _ipopt_options[b"print_level"] > 12:  # type: ignore
        for key in ["fun", "message", "nfev", "nit", "njev", "status", "success"]:
            print(key + ":", result[key])
        X = model_formula.get_model_matrix(design).to_numpy()
        print("metrics:", metrics(X))

    # check if all points respect the domain and the constraint
    try:
        domain.validate_candidates(
            candidates=design.apply(lambda x: np.round(x, 8)),
            only_inputs=True,
            tol=1e-4,
        )
    except (ValueError, ConstraintNotFulfilledError):
        warnings.warn(
            "Some points do not lie inside the domain or violate constraints. Please check if the \
                results lie within your tolerance.",
            UserWarning,
        )

    return design


def partially_fix_experiment(
    bounds: list,
    fixed_experiments: Union[pd.DataFrame, None],
    n_experiments: int,
    partially_fixed_experiments: Union[pd.DataFrame, None],
    x0: np.ndarray,
) -> Tuple[List, np.ndarray]:
    """
    fixes some variables for experiments. Within one experiment not all variables need to be fixed.
    Variables can be fixed to one value or can be set to a range by setting a tuple with lower and upper bound
    Non-fixed variables have to be set to None or nan. Will also fix the experiments provided in fixed_experiments

    Args:
        bounds (list): current bounds
        fixed_experiments (pd.Dataframe): experiments which are guaranteed to be part of the design and are fully fixed
        n_experiments (int): number of experiments
        partially_fixed_experiments (pd.Dataframe): experiments which are partially fixed
        x0: initial guess

    Returns: Tuple of list and pd.Dataframe containing the new bounds for each variable and an adapted initial guess
        which comply with the bounds

    """

    shift = 0
    if partially_fixed_experiments is not None:
        partially_fixed_experiments.sort_index(axis=1, inplace=True)
        if fixed_experiments is not None:
            if (
                len(fixed_experiments) + len(partially_fixed_experiments)
                > n_experiments
            ):
                raise AttributeError(
                    "Number of fixed experiments and partially fixed experiments exceeds the number of total "
                    "experiments"
                )
            shift = len(fixed_experiments)

        shift = shift * len(partially_fixed_experiments.columns)
        for i, val in enumerate(np.array(partially_fixed_experiments.values).flatten()):
            index = shift + i
            if type(val) is tuple:
                bounds[index] = (val[0], val[1])
                x0[index] = val[0]
            elif val is not None and not np.isnan(val):
                bounds[index] = (val, val)
                x0[index] = val
    return bounds, x0


def check_fixed_experiments(
    domain: Domain, n_experiments: int, fixed_experiments: np.ndarray
) -> None:
    """Checks if the shape of the fixed experiments is correct and if the number of fixed experiments is valid
    Args:
        domain (Domain): domain defining the input variables used for the check.
        n_experiments (int): total number of experiments in the design that fixed_experiments are part of.
        fixed_experiments (np.ndarray): fixed experiment proposals to be checked.
    """

    n_fixed_experiments, D = np.array(fixed_experiments).shape

    if n_fixed_experiments >= n_experiments:
        raise ValueError(
            "For starting the optimization the total number of experiments must be larger that the number of fixed experiments."
        )

    if D != len(domain.inputs):
        raise ValueError(
            f"Invalid shape of fixed_experiments. Length along axis 1 is {D}, but must be {len(domain.inputs)}"
        )


def check_partially_and_fully_fixed_experiments(
    domain: Domain,
    n_experiments: int,
    fixed_experiments: np.ndarray,
    paritally_fixed_experiments: np.ndarray,
) -> None:
    """Checks if the shape of the fixed experiments is correct and if the number of fixed experiments is valid
    Args:
        domain (Domain): domain defining the input variables used for the check.
        n_experiments (int): total number of experiments in the design that fixed_experiments are part of.
        fixed_experiments (np.ndarray): fixed experiment proposals to be checked.
        paritally_fixed_experiments (np.ndarray): partially fixed experiment proposals to be checked.
    """

    check_fixed_experiments(domain, n_experiments, fixed_experiments)
    n_fixed_experiments, dim = np.array(fixed_experiments).shape

    n_partially_fixed_experiments, partially_dim = np.array(
        paritally_fixed_experiments
    ).shape

    if partially_dim != len(domain.inputs):
        raise ValueError(
            f"Invalid shape of partially_fixed_experiments. Length along axis 1 is {partially_dim}, but must be {len(domain.inputs)}"
        )

    if n_fixed_experiments + n_partially_fixed_experiments > n_experiments:
        warnings.warn(
            UserWarning(
                "The number of fixed experiments and partially fixed experiments exceeds the amount "
                "of the overall count of experiments. Partially fixed experiments may be cut of"
            )
        )


def get_n_experiments(
    domain: Domain, model_type: Union[str, Formula], n_experiments: Optional[int] = None
):
    """Determines a number of experiments which is appropriate for the model if no
    number is provided. Otherwise warns if the provided number of experiments is smaller than recommended.

    Args:
        domain (Domain): domain containing the model inputs.
        model_type (str, Formula): keyword or formulaic Formula describing the model.
        n_experiments (int, optional): number of experiments. Defaults to zero.

    Returns:
        n_experiments if an integer value for n_experiments is given. Number of model terms + 3 otherwise.

    """
    n_experiments_min = (
        len(
            get_formula_from_string(model_type=model_type, rhs_only=True, domain=domain)
        )
        + 3
    )

    if n_experiments is None:
        n_experiments = n_experiments_min
    elif n_experiments < n_experiments_min:
        warnings.warn(
            f"The minimum number of experiments is {n_experiments_min}, but the current setting is n_experiments={n_experiments}."
        )

    return n_experiments



import numpy as np
import pandas as pd

from bofire.data_models.constraints.api import NonlinearInequalityConstraint


def test_nonlinear_constraints_jacobian_expression():
    constraint0 = NonlinearInequalityConstraint(
        expression="x1**2 + x2**2 - x3", features=["x1", "x2", "x3"]
    )
    constraint1 = NonlinearInequalityConstraint(
        expression="x1**2 + x2**2 - x3",
        features=["x1", "x2", "x3"],
        jacobian_expression="[2*x1, 2*x2, -1]",
    )

    data = pd.DataFrame(np.random.rand(10, 3), columns=["x1", "x2", "x3"])
    assert np.allclose(constraint0.jacobian(data), constraint1.jacobian(data))



import itertools
from typing import List, Union

from pydantic import validator

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import TInputTransformSpecs
from bofire.data_models.surrogates.empirical import EmpiricalSurrogate
from bofire.data_models.surrogates.fully_bayesian import SaasSingleTaskGPSurrogate
from bofire.data_models.surrogates.linear import LinearSurrogate
from bofire.data_models.surrogates.mixed_single_task_gp import (
    MixedSingleTaskGPSurrogate,
)
from bofire.data_models.surrogates.mlp import MLPEnsemble
from bofire.data_models.surrogates.polynomial import PolynomialSurrogate
from bofire.data_models.surrogates.random_forest import RandomForestSurrogate
from bofire.data_models.surrogates.single_task_gp import SingleTaskGPSurrogate
from bofire.data_models.surrogates.tanimoto_gp import TanimotoGPSurrogate

AnyBotorchSurrogate = Union[
    EmpiricalSurrogate,
    RandomForestSurrogate,
    SingleTaskGPSurrogate,
    MixedSingleTaskGPSurrogate,
    MLPEnsemble,
    SaasSingleTaskGPSurrogate,
    TanimotoGPSurrogate,
    LinearSurrogate,
    PolynomialSurrogate,
]


class BotorchSurrogates(BaseModel):
    """ "List of botorch surrogates.

    Behaves similar to a Surrogate."""

    surrogates: List[AnyBotorchSurrogate]

    @property
    def input_preprocessing_specs(self) -> TInputTransformSpecs:
        return {
            key: value
            for model in self.surrogates
            for key, value in model.input_preprocessing_specs.items()
        }

    @property
    def outputs(self) -> Outputs:
        return Outputs(
            features=list(
                itertools.chain.from_iterable(
                    [model.outputs.get() for model in self.surrogates]  # type: ignore
                )
            )
        )

    def _check_compability(self, inputs: Inputs, outputs: Outputs):
        # TODO: add sync option
        used_output_feature_keys = self.outputs.get_keys()
        if sorted(used_output_feature_keys) != sorted(outputs.get_keys()):
            raise ValueError("Output features do not match.")
        used_feature_keys = []
        for i, model in enumerate(self.surrogates):
            if len(model.inputs) > len(inputs):
                raise ValueError(
                    f"Model with index {i} has more features than acceptable."
                )
            for feat in model.inputs:
                try:
                    other_feat = inputs.get_by_key(feat.key)
                except KeyError:
                    raise ValueError(f"Feature {feat.key} not found.")
                # now compare the features
                # TODO: make more sohisticated comparisons based on the type
                # has to to be implemented in features, for the start
                # we go with __eq__
                if feat != other_feat:
                    raise ValueError(f"Features with key {feat.key} are incompatible.")
                if feat.key not in used_feature_keys:
                    used_feature_keys.append(feat.key)
        if len(used_feature_keys) != len(inputs):
            raise ValueError("Unused features are present.")

    @validator("surrogates")
    def validate_surrogates(cls, v, values):
        # validate that all surrogates are single output surrogates
        # TODO: this restriction has to be removed at some point
        for model in v:
            if len(model.outputs) != 1:
                raise ValueError("Only single output surrogates allowed.")
        # check that the output feature keys are distinctw
        used_output_feature_keys = list(
            itertools.chain.from_iterable([model.outputs.get_keys() for model in v])
        )
        if len(set(used_output_feature_keys)) != len(used_output_feature_keys):
            raise ValueError("Output feature keys are not unique across surrogates.")
        # get the feature keys present in all surrogates
        used_feature_keys = []
        for model in v:
            for key in model.inputs.get_keys():
                if key not in used_feature_keys:
                    used_feature_keys.append(key)
        # check that the features and preprocessing steps are equal trough the surrogates
        for key in used_feature_keys:
            features = [
                model.inputs.get_by_key(key)
                for model in v
                if key in model.inputs.get_keys()
            ]
            preproccessing = [
                model.input_preprocessing_specs[key]
                for model in v
                if key in model.input_preprocessing_specs
            ]
            if all(features) is False:
                raise ValueError(f"Features with key {key} are incompatible.")
            if all(i == preproccessing[0] for i in preproccessing) is False:
                raise ValueError(
                    f"Preprocessing steps for features with {key} are incompatible."
                )
        return v



from bofire.strategies.doe_strategy import DoEStrategy  # noqa: F401
from bofire.strategies.mapper import map  # noqa: F401
from bofire.strategies.predictives.botorch import BotorchStrategy  # noqa: F401
from bofire.strategies.predictives.predictive import PredictiveStrategy  # noqa: F401
from bofire.strategies.predictives.qehvi import QehviStrategy  # noqa: F401
from bofire.strategies.predictives.qnehvi import QnehviStrategy  # noqa: F401
from bofire.strategies.predictives.qparego import QparegoStrategy  # noqa: F401
from bofire.strategies.predictives.sobo import (  # noqa: F401
    AdditiveSoboStrategy,
    CustomSoboStrategy,
    MultiplicativeSoboStrategy,
    SoboStrategy,
)
from bofire.strategies.random import RandomStrategy  # noqa: F401
from bofire.strategies.samplers.polytope import PolytopeSampler  # noqa: F401
from bofire.strategies.samplers.rejection import RejectionSampler  # noqa: F401
from bofire.strategies.samplers.sampler import SamplerStrategy  # noqa: F401
from bofire.strategies.stepwise.stepwise import StepwiseStrategy  # noqa: F401
from bofire.strategies.strategy import Strategy  # noqa: F401



import numpy as np
import pandas as pd
import pytest

import tests.bofire.data_models.specs.api as specs
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
    MinimizeSigmoidObjective,
)
from bofire.utils.multiobjective import (
    compute_hypervolume,
    get_pareto_front,
    get_ref_point_mask,
    infer_ref_point,
)

if1 = specs.features.valid(ContinuousInput).obj(key="if1")
if2 = specs.features.valid(ContinuousInput).obj(key="if2")

of1 = ContinuousOutput(
    objective=MaximizeObjective(w=1),
    key="of1",
)
of2 = ContinuousOutput(
    objective=MinimizeObjective(w=1),
    key="of2",
)
of3 = ContinuousOutput(
    objective=MaximizeObjective(w=1),
    key="of3",
)
of4 = ContinuousOutput(
    objective=MinimizeObjective(w=1),
    key="of4",
)
of5 = ContinuousOutput(
    objective=MaximizeSigmoidObjective(w=1, tp=1, steepness=1),
    key="of5",
)
of6 = ContinuousOutput(
    objective=MinimizeSigmoidObjective(w=1, tp=1, steepness=1),
    key="of6",
)
of7 = ContinuousOutput(
    objective=CloseToTargetObjective(w=1, target_value=5, exponent=1),
    key="of7",
)

valid_domains = [
    Domain(inputs=[if1, if2], outputs=[of1, of2]),
    Domain(inputs=[if1, if2], outputs=[of1, of3]),
    Domain(inputs=[if1, if2], outputs=[of1, of2, of3, of4]),
    Domain(inputs=[if1, if2], outputs=[of2, of4]),
    Domain(inputs=[if1, if2], outputs=[of2, of1, of3, of4]),
    Domain(inputs=[if1, if2], outputs=[of1, of2, of7]),
]

valid_constrained_domains = [
    Domain(inputs=[if1, if2], outputs=[of1, of2, of5]),
    Domain(inputs=[if1, if2], outputs=[of1, of2, of6]),
    Domain(inputs=[if1, if2], outputs=[of1, of2, of5, of6]),
]

invalid_domains = [
    Domain(inputs=[if1, if2], outputs=[of1]),
    Domain(inputs=[if1, if2], outputs=[of1, of5]),
]

dfs = [
    pd.DataFrame.from_dict(
        {
            "if1": [3.0, 4.0, 5.0, 6.0],
            "if2": [10.0, 7.0, 8.0, 12.0],
            "of1": [
                1.0,
                10.0,
                4.0,
                5.0,
            ],
            "of2": [
                5.0,
                3.0,
                2.0,
                5.0,
            ],
            "valid_of1": [
                1,
                1,
                1,
                1,
            ],
            "valid_of2": [
                1,
                1,
                1,
                1,
            ],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3.0, 4.0, 5.0, 6.0],
            "if2": [10.0, 7.0, 8.0, 12.0],
            "of1": [
                1.0,
                10.0,
                4.0,
                5.0,
            ],
            "of3": [
                5.0,
                3.0,
                2.0,
                5.0,
            ],
            "valid_of1": [
                1,
                1,
                1,
                1,
            ],
            "valid_of3": [
                1,
                1,
                1,
                1,
            ],
        }
    ),
    pd.DataFrame.from_dict(
        {
            "if1": [3.0, 4.0, 5.0, 6.0],
            "if2": [10.0, 7.0, 8.0, 12.0],
            "of1": [
                1.0,
                10.0,
                4.0,
                5.0,
            ],
            "of2": [
                5.0,
                3.0,
                2.0,
                5.0,
            ],
            "of7": [
                10.0,
                0.0,
                30.0,
                6.0,
            ],
            "valid_of1": [
                1,
                1,
                1,
                1,
            ],
            "valid_of2": [
                1,
                1,
                1,
                1,
            ],
            "valid_of7": [
                1,
                1,
                1,
                1,
            ],
        }
    ),
]


@pytest.mark.parametrize(
    "domain, expected",
    [
        (valid_domains[0], np.array([1.0, -1.0])),
        (valid_constrained_domains[0], np.array([1.0, -1.0])),
        (valid_domains[1], np.array([1.0, 1.0])),
        (valid_domains[2], np.array([1.0, -1.0, 1.0, -1.0])),
        (valid_domains[3], np.array([-1.0, -1.0])),
        (valid_domains[4], np.array([1.0, -1.0, 1.0, -1.0])),
        (valid_domains[5], np.array([1.0, -1.0, -1.0])),
    ],
)
def test_get_ref_point_mask(domain, expected):
    assert np.allclose(get_ref_point_mask(domain), expected)


@pytest.mark.parametrize(
    "domain, subset, expected",
    [
        (valid_domains[2], ["of1", "of2"], np.array([1.0, -1.0])),
        (valid_domains[2], ["of1", "of2", "of3"], np.array([1.0, -1.0, 1.0])),
        (valid_domains[2], ["of1", "of3"], np.array([1.0, 1.0])),
        (valid_domains[5], ["of1", "of7"], np.array([1.0, -1.0])),
    ],
)
def test_get_ref_point_mask_subset(domain, subset, expected):
    assert np.allclose(get_ref_point_mask(domain, output_feature_keys=subset), expected)


@pytest.mark.parametrize("domain", invalid_domains)
def test_invalid_get_ref_point_mask(domain):
    with pytest.raises(ValueError):
        get_ref_point_mask(domain)


@pytest.mark.parametrize(
    "domain, experiments, expected_indices",
    [
        (valid_domains[0], dfs[0], np.array([1, 2], dtype="int64")),
        (valid_constrained_domains[0], dfs[0], np.array([1, 2], dtype="int64")),
        (valid_domains[1], dfs[1], np.array([1, 3], dtype="int64")),
        (valid_domains[5], dfs[2], np.array([1, 2, 3], dtype="int64")),
    ],
)
def test_get_pareto_front(domain, experiments, expected_indices):
    df_pareto = get_pareto_front(domain, experiments)
    assert np.allclose(df_pareto.index.values, expected_indices)


@pytest.mark.parametrize(
    "domain, experiments, ref_point",
    [
        (valid_domains[0], dfs[0], {"of1": 0.0, "of2": 20.0}),
        (valid_constrained_domains[0], dfs[0], {"of1": 0.0, "of2": 20.0}),
        (valid_domains[1], dfs[1], {"of1": 0.0, "of3": 0.0}),
        (valid_domains[5], dfs[2], {"of1": 0.0, "of2": 20.0, "of7": 25.0}),
    ],
)
def test_compute_hypervolume(domain, experiments, ref_point):
    df_pareto = get_pareto_front(domain, experiments)
    hv = compute_hypervolume(domain, df_pareto, ref_point)
    assert hv > 0


@pytest.mark.parametrize(
    "domain, experiments, return_masked, expected",
    [
        (valid_domains[0], dfs[0], True, {"of1": 1.0, "of2": -5.0}),
        (valid_domains[0], dfs[0], False, {"of1": 1.0, "of2": 5.0}),
        (valid_domains[1], dfs[1], True, {"of1": 1.0, "of3": 2.0}),
        (valid_domains[1], dfs[1], False, {"of1": 1.0, "of3": 2.0}),
        (valid_constrained_domains[0], dfs[0], True, {"of1": 1.0, "of2": -5.0}),
        (valid_constrained_domains[0], dfs[0], False, {"of1": 1.0, "of2": 5.0}),
        (valid_domains[5], dfs[2], True, {"of1": 1.0, "of2": -5.0, "of7": -25.0}),
        (valid_domains[5], dfs[2], False, {"of1": 1.0, "of2": 5.0, "of7": 25.0}),
    ],
)
def test_infer_ref_point(domain, experiments, return_masked, expected):
    ref_point = infer_ref_point(domain, experiments, return_masked)
    keys = domain.outputs.get_keys_by_objective(
        includes=[MaximizeObjective, MinimizeObjective]
    )
    assert np.allclose(
        np.array([ref_point[feat] for feat in keys]),
        np.array([expected[feat] for feat in keys]),
    )



from typing import Dict, Type

from bofire.data_models.outlier_detection import api as data_models
from bofire.outlier_detection.outlier_detection import (
    IterativeTrimming,
    OutlierDetection,
)

OUTLIER_MAP: Dict[Type[data_models.OutlierDetection], Type[OutlierDetection]] = {
    data_models.IterativeTrimming: IterativeTrimming,
}


def map(data_model: data_models.OutlierDetection) -> OutlierDetection:
    cls = OUTLIER_MAP[data_model.__class__]
    return cls(data_model=data_model)  # type: ignore



from abc import abstractmethod
from typing import Type

from pydantic import validator

from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import Output
from bofire.data_models.objectives.api import Objective
from bofire.data_models.strategies.strategy import Strategy


class PredictiveStrategy(Strategy):
    @validator("domain")
    def validate_objectives(cls, domain: Domain):
        """Validator to ensure that all objectives defined in the domain are valid for the chosen strategy

        Args:
            domain (Domain): The domain to be used in the strategy

        Raises:
            ValueError: if a objective type is defined in the domain but is invalid for the strategy chosen

        Returns:
            Domain: the domain
        """
        for feature in domain.outputs.get_by_objective(Objective):
            assert isinstance(feature, Output)
            assert feature.objective is not None
            if not cls.is_objective_implemented(type(feature.objective)):  # type: ignore
                raise ValueError(
                    f"Objective `{type(feature.objective)}` is not implemented for strategy `{cls.__name__}`"  # type: ignore
                )
        return domain

    @classmethod
    @abstractmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        """Abstract method to check if a objective type is implemented for the strategy

        Args:
            my_type (Type[Objective]): Objective class

        Returns:
            bool: True if the objective type is valid for the strategy chosen, False otherwise
        """
        pass

    @validator("domain")
    def validate_output_feature_count(cls, domain: Domain):
        """Validator to ensure that at least one output feature with attached objective is defined.

        Args:
            domain (Domain): The domain to be used in the strategy

        Raises:
            ValueError: if no output feature is specified
            ValueError: if not output feauture with an attached objective is specified

        Returns:
            Domain: the domain
        """
        if len(domain.outputs) == 0:
            raise ValueError("no output feature specified")
        if len(domain.outputs.get_by_objective(Objective)) == 0:
            raise ValueError("no output feature with objective specified")
        return domain



import warnings

import pytest

import bofire.data_models.strategies.api as data_models
import bofire.strategies.api as strategies
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)
from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    DiscreteInput,
)
from bofire.strategies.api import RandomStrategy

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning, append=True)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

if0 = ContinuousInput(key="if0", bounds=(0, 1))
if1 = ContinuousInput(key="if1", bounds=(0, 2))
if2 = ContinuousInput(key="if2", bounds=(0, 3))
if3 = CategoricalInput(key="if3", categories=["c1", "c2", "c3"])
if4 = CategoricalInput(
    key="if4", categories=["A", "B", "C"], allowed=[True, True, False]
)
if5 = CategoricalInput(key="if5", categories=["A", "B"], allowed=[True, False])
if6 = CategoricalDescriptorInput(
    key="if6",
    categories=["A", "B", "C"],
    descriptors=["d1", "d2"],
    values=[[1, 2], [3, 7], [5, 1]],
)
if7 = DiscreteInput(key="if7", values=[0, 1, 5])

of1 = ContinuousOutput(key="of1")

c1 = LinearEqualityConstraint(features=["if0", "if1"], coefficients=[1, 1], rhs=1)
c2 = LinearInequalityConstraint(features=["if0", "if1"], coefficients=[1, 1], rhs=1)
c3 = NonlinearEqualityConstraint(expression="if0**2 + if1**2 - 1")
c4 = NonlinearInequalityConstraint(expression="if0**2 + if1**2 - 1")
c5 = NChooseKConstraint(
    features=["if0", "if1", "if2"], min_count=0, max_count=2, none_also_valid=False
)

supported_domains = [
    Domain.from_lists(
        # continuous features
        inputs=[if0, if1],
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        # continuous features incl. with fixed values
        inputs=[if0, if1, if2],
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        # all feature types
        inputs=[if1, if3, if6, if7],
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        # all feature types incl. with fixed values
        inputs=[if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[],
    ),
    Domain.from_lists(
        # all feature types, linear equality
        inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[c1],
    ),
    Domain.from_lists(
        # all feature types, linear inequality
        inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[c2],
    ),
    Domain.from_lists(
        # all feature types, nonlinear inequality
        inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[c4],
    ),
]

unsupported_domains = [
    Domain.from_lists(
        # nonlinear equality
        inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[c3],
    ),
    Domain.from_lists(
        # combination of linear equality and nonlinear inequality
        inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
        outputs=[of1],
        constraints=[c1, c4],
    ),
    # Domain(
    #     # n-choose-k
    #     inputs=[if0, if1, if2, if3, if4, if5, if6, if7],
    #     outputs=[of1],
    #     constraints=[c5],
    # ),
]


@pytest.mark.parametrize("domain", supported_domains)
def test_ask(domain):
    data_model = data_models.RandomStrategy(domain=domain)
    strategy = strategies.map(data_model=data_model)
    candidates = strategy.ask(3)
    assert len(candidates) == 3


@pytest.mark.parametrize("domain", unsupported_domains)
def test_unsupported(domain):
    with pytest.raises(Exception):
        data_model = data_models.RandomStrategy(domain=domain)
        RandomStrategy(data_model=data_model)



from typing import Literal, Type, Union

from pydantic import Field

from bofire.data_models.acquisition_functions.api import qEI, qLogEI, qLogNEI, qNEI
from bofire.data_models.features.api import CategoricalOutput, Feature
from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
    MinimizeSigmoidObjective,
    Objective,
    TargetObjective,
)
from bofire.data_models.strategies.predictives.multiobjective import (
    MultiobjectiveStrategy,
)


class QparegoStrategy(MultiobjectiveStrategy):
    type: Literal["QparegoStrategy"] = "QparegoStrategy"
    acquisition_function: Union[qEI, qLogEI, qLogNEI, qNEI] = Field(
        default_factory=lambda: qNEI()
    )

    @classmethod
    def is_feature_implemented(cls, my_type: Type[Feature]) -> bool:
        if my_type not in [CategoricalOutput]:
            return True
        return False

    @classmethod
    def is_objective_implemented(cls, my_type: Type[Objective]) -> bool:
        if my_type not in [
            MaximizeObjective,
            MinimizeObjective,
            TargetObjective,
            MinimizeSigmoidObjective,
            MaximizeSigmoidObjective,
            CloseToTargetObjective,
        ]:
            return False
        return True



import pytest

import bofire.data_models.strategies.api as data_models
import bofire.data_models.surrogates.api as surrogate_data_models
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.strategies.strategy import Strategy
from tests.bofire.strategies.specs import (
    VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
    VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
)
from tests.bofire.strategies.test_qehvi import VALID_BOTORCH_QEHVI_STRATEGY_SPEC

if1 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if11",
    }
)
if2 = ContinuousInput(
    **{
        **VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
        "key": "if22",
    }
)
of1 = ContinuousOutput(**{**VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC, "key": "of11"})
of2 = ContinuousOutput(**{**VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC, "key": "of22"})


@pytest.mark.parametrize(
    "strategy, expected_count",
    [
        (
            data_models.QnehviStrategy(
                **VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
            ),
            2,
        ),
        (
            data_models.QnehviStrategy(
                **{
                    **VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
                    "surrogate_specs": surrogate_data_models.BotorchSurrogates(
                        surrogates=[
                            surrogate_data_models.SingleTaskGPSurrogate(
                                inputs=VALID_BOTORCH_QEHVI_STRATEGY_SPEC[
                                    "domain"
                                ].inputs,
                                outputs=Outputs(
                                    features=[
                                        VALID_BOTORCH_QEHVI_STRATEGY_SPEC[
                                            "domain"
                                        ].outputs.get_by_key("of1")
                                    ]
                                ),
                            ),
                        ]
                    ),
                },
            ),
            2,
        ),
    ],
)
def test_generate_surrogate_specs(strategy: Strategy, expected_count: int):
    surrogate_specs = data_models.BotorchStrategy._generate_surrogate_specs(
        domain=strategy.domain
    )  # , surrogate_specs=strategy.surrogate_specs
    # )
    assert len(surrogate_specs.surrogates) == expected_count


@pytest.mark.parametrize(
    "strategy, specs",
    [
        (
            data_models.QnehviStrategy(
                **VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
            ),
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=Inputs(features=[if1, if2]),
                        outputs=Outputs(
                            features=[
                                VALID_BOTORCH_QEHVI_STRATEGY_SPEC[
                                    "domain"
                                ].outputs.get_by_key("of1")
                            ]
                        ),
                    ),
                ]
            ),
        ),
        (
            data_models.QnehviStrategy(
                **VALID_BOTORCH_QEHVI_STRATEGY_SPEC,
            ),
            surrogate_data_models.BotorchSurrogates(
                surrogates=[
                    surrogate_data_models.SingleTaskGPSurrogate(
                        inputs=VALID_BOTORCH_QEHVI_STRATEGY_SPEC["domain"].inputs,
                        outputs=Outputs(features=[of1]),
                    ),
                ]
            ),
        ),
    ],
)
def test_generate_surrogate_specs_invalid(
    strategy: data_models.Strategy, specs: surrogate_data_models.BotorchSurrogates
):
    with pytest.raises(ValueError):
        data_models.BotorchStrategy._generate_surrogate_specs(strategy.domain, specs)



from bofire.surrogates.botorch_surrogates import BotorchSurrogates
from bofire.surrogates.empirical import EmpiricalSurrogate
from bofire.surrogates.mapper import map
from bofire.surrogates.mixed_single_task_gp import MixedSingleTaskGPSurrogate
from bofire.surrogates.mlp import MLPEnsemble
from bofire.surrogates.random_forest import RandomForestSurrogate
from bofire.surrogates.single_task_gp import SingleTaskGPSurrogate
from bofire.surrogates.surrogate import Surrogate
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.surrogates.values import PredictedValue
from bofire.surrogates.xgb import XGBoostSurrogate



import pandas as pd
import pytest

import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate


@pytest.mark.parametrize("folds", [5, 3, 10, -1])
def test_model_cross_validate(folds):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=100)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    experiments = experiments.sample(10)
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    train_cv, test_cv, _ = model.cross_validate(experiments, folds=folds)
    efolds = folds if folds != -1 else 10
    assert len(train_cv.results) == efolds
    assert len(test_cv.results) == efolds


def test_model_cross_validate_descriptor():
    folds = 5
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [
            CategoricalDescriptorInput(
                key="x_3",
                categories=["a", "b", "c"],
                descriptors=["alpha"],
                values=[[1], [2], [3]],
            )
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=100)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments.loc[experiments.x_2 == "b", "y"] += 5
    experiments.loc[experiments.x_2 == "c", "y"] += 10
    experiments["valid_y"] = 1
    experiments = experiments.sample(10)
    for encoding in [
        CategoricalEncodingEnum.ONE_HOT,
        CategoricalEncodingEnum.DESCRIPTOR,
    ]:
        model = SingleTaskGPSurrogate(
            inputs=inputs,
            outputs=outputs,
            input_preprocessing_specs={"x_3": encoding},
        )
        model = surrogates.map(model)
        train_cv, test_cv, _ = model.cross_validate(experiments, folds=folds)
        efolds = folds if folds != -1 else 10
        assert len(train_cv.results) == efolds
        assert len(test_cv.results) == efolds


@pytest.mark.parametrize("include_X, include_labcodes", [[True, False], [False, True]])
def test_model_cross_validate_include_X(include_X, include_labcodes):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments["labcode"] = [str(i) for i in range(10)]
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    train_cv, test_cv, _ = model.cross_validate(
        experiments, folds=5, include_X=include_X, include_labcodes=include_labcodes
    )
    if include_X:
        assert train_cv.results[0].X.shape == (8, 2)
        assert test_cv.results[0].X.shape == (2, 2)
    if include_X is False:
        assert train_cv.results[0].X is None
        assert test_cv.results[0].X is None
    if include_labcodes:
        assert train_cv.results[0].labcodes.shape == (8,)
        assert test_cv.results[0].labcodes.shape == (2,)
    else:
        assert train_cv.results[0].labcodes is None
        assert train_cv.results[0].labcodes is None


def test_model_cross_validate_hooks():
    def hook1(surrogate, X_train, y_train, X_test, y_test):
        assert isinstance(surrogate, surrogates.SingleTaskGPSurrogate)
        assert y_train.shape == (8, 1)
        assert y_test.shape == (2, 1)
        return X_train.shape

    def hook2(surrogate, X_train, y_train, X_test, y_test, return_test=True):
        if return_test:
            return X_test.shape
        return X_train.shape

    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    #
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    # first test with one hook
    _, _, hook_results = model.cross_validate(
        experiments, folds=5, hooks={"hook1": hook1}
    )
    assert len(hook_results.keys()) == 1
    assert len(hook_results["hook1"]) == 5
    assert hook_results["hook1"] == [(8, 2), (8, 2), (8, 2), (8, 2), (8, 2)]
    # now test with two hooks
    _, _, hook_results = model.cross_validate(
        experiments, folds=5, hooks={"hook1": hook1, "hook2": hook2}
    )
    assert len(hook_results.keys()) == 2
    assert len(hook_results["hook1"]) == 5
    assert hook_results["hook1"] == [(8, 2), (8, 2), (8, 2), (8, 2), (8, 2)]
    assert len(hook_results["hook2"]) == 5
    assert hook_results["hook2"] == [(2, 2), (2, 2), (2, 2), (2, 2), (2, 2)]
    # now test with two hooks and keyword arguments
    _, _, hook_results = model.cross_validate(
        experiments,
        folds=5,
        hooks={"hook1": hook1, "hook2": hook2},
        hook_kwargs={"hook2": {"return_test": False}},
    )
    assert len(hook_results.keys()) == 2
    assert len(hook_results["hook1"]) == 5
    assert hook_results["hook1"] == [(8, 2), (8, 2), (8, 2), (8, 2), (8, 2)]
    assert len(hook_results["hook2"]) == 5
    assert hook_results["hook2"] == [(8, 2), (8, 2), (8, 2), (8, 2), (8, 2)]


@pytest.mark.parametrize("folds", [-2, 0, 1])
def test_model_cross_validate_invalid(folds):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    with pytest.raises(ValueError):
        model.cross_validate(experiments, folds=folds)


@pytest.mark.parametrize("folds", [5, 3, 10, -1])
def test_model_cross_validate_random_state(folds):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=100)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    experiments = experiments.sample(10)
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    train_cv_1_1, test_cv_1_1, _ = model.cross_validate(
        experiments, folds=folds, random_state=1
    )
    train_cv_1_2, test_cv_1_2, _ = model.cross_validate(
        experiments, folds=folds, random_state=1
    )

    train_cv_2_1, test_cv_2_1, _ = model.cross_validate(
        experiments, folds=folds, random_state=2
    )
    train_cv_2_2, test_cv_2_2, _ = model.cross_validate(
        experiments, folds=folds, random_state=2
    )

    for cvresult1, cvresult2 in zip(train_cv_1_1.results, train_cv_1_2.results):
        assert all(list(cvresult1.observed.index == cvresult2.observed.index))

    for cvresult1, cvresult2 in zip(test_cv_1_1.results, test_cv_1_2.results):
        assert all(list(cvresult1.observed.index == cvresult2.observed.index))

    for cvresult1, cvresult2 in zip(train_cv_2_1.results, train_cv_2_2.results):
        assert all(list(cvresult1.observed.index == cvresult2.observed.index))

    for cvresult1, cvresult2 in zip(test_cv_2_1.results, test_cv_2_2.results):
        assert all(list(cvresult1.observed.index == cvresult2.observed.index))

    for cvresult1, cvresult2 in zip(test_cv_1_1.results, test_cv_2_1.results):
        assert not all(list(cvresult1.observed.index == cvresult2.observed.index))

    for cvresult1, cvresult2 in zip(test_cv_1_2.results, test_cv_2_2.results):
        assert not all(list(cvresult1.observed.index == cvresult2.observed.index))


# Include test for CategoricalOutput when fully implemented
@pytest.mark.parametrize("random_state", [1, 2])
def test_model_cross_validate_stratified(random_state):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [
            CategoricalInput(key="cat_x_3", categories=["category1", "category2"]),
            CategoricalDescriptorInput(
                key="cat_x_4",
                categories=["a", "b", "c"],
                descriptors=["alpha"],
                values=[[1], [2], [3]],
            ),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    # category2, b, and c only appears 5 times each
    experiments = pd.DataFrame(
        [
            [-4, -4, "category1", "a", 1],
            [-3, -3, "category1", "a", 1],
            [-2, -2, "category1", "a", 1],
            [-1, -1, "category1", "b", 1],
            [0, 0, "category1", "b", 1],
            [1, 1, "category1", "b", 1],
            [2, 2, "category1", "c", 1],
            [3, 3, "category1", "c", 1],
            [2, 3, "category1", "c", 1],
            [3, 1, "category1", "a", 1],
            [3, 4, "category1", "a", 0],
            [4, 4, "category2", "b", 0],
            [1, 4, "category2", "b", 0],
            [1, 0, "category2", "c", 0],
            [1, 2, "category2", "c", 0],
            [2, 4, "category2", "a", 1],
        ],
        columns=["x_1", "x_2", "cat_x_3", "cat_x_4", "y"],
    )
    experiments["valid_y"] = 1

    cat_x_3_category2_indexes = experiments.index[
        experiments["cat_x_3"] == "category2"
    ].tolist()
    cat_x_4_b_indexes = experiments.index[experiments["cat_x_4"] == "b"].tolist()
    cat_x_4_c_indexes = experiments.index[experiments["cat_x_4"] == "c"].tolist()
    zero_indexes = experiments.index[experiments["y"] == 0].tolist()

    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    _, test_cv, _ = model.cross_validate(
        experiments, folds=5, random_state=random_state, stratified_feature="cat_x_3"
    )
    for cvresults in test_cv.results:
        assert any(i in cvresults.observed.index for i in cat_x_3_category2_indexes)

    _, test_cv, _ = model.cross_validate(
        experiments, folds=5, random_state=random_state, stratified_feature="cat_x_4"
    )
    for cvresults in test_cv.results:
        assert any(i in cvresults.observed.index for i in cat_x_4_b_indexes)
        assert any(i in cvresults.observed.index for i in cat_x_4_c_indexes)

    _, test_cv, _ = model.cross_validate(
        experiments, folds=5, random_state=random_state, stratified_feature="y"
    )
    for cvresults in test_cv.results:
        assert any(i in cvresults.observed.index for i in zero_indexes)


def test_model_cross_validate_stratified_invalid_feature_name():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [
            CategoricalInput(key="cat_x_3", categories=["category1", "category2"]),
            CategoricalDescriptorInput(
                key="cat_x_4",
                categories=["a", "b", "c"],
                descriptors=["alpha"],
                values=[[1], [2], [3]],
            ),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = pd.DataFrame(
        [
            [-4, -4, "category1", "a", 1],
            [-3, -3, "category1", "a", 1],
            [-2, -2, "category1", "a", 1],
            [-1, -1, "category1", "b", 1],
            [0, 0, "category1", "b", 1],
            [1, 1, "category1", "b", 1],
            [2, 2, "category1", "c", 1],
            [3, 3, "category1", "c", 1],
            [2, 3, "category1", "c", 1],
            [3, 1, "category1", "a", 1],
            [3, 4, "category1", "a", 0],
            [4, 4, "category2", "b", 0],
            [1, 4, "category2", "b", 0],
            [1, 0, "category2", "c", 0],
            [1, 2, "category2", "c", 0],
            [2, 4, "category2", "a", 1],
        ],
        columns=["x_1", "x_2", "cat_x_3", "cat_x_4", "y"],
    )
    experiments["valid_y"] = 1

    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    with pytest.raises(
        ValueError,
        match="The feature to be stratified is not in the model inputs or outputs",
    ):
        model.cross_validate(experiments, folds=5, stratified_feature="name")


@pytest.mark.parametrize("key", ["x_1", "x_2"])
def test_model_cross_validate_stratified_invalid_feature_type(key):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [
            CategoricalInput(key="cat_x_3", categories=["category1", "category2"]),
            CategoricalDescriptorInput(
                key="cat_x_4",
                categories=["a", "b", "c"],
                descriptors=["alpha"],
                values=[[1], [2], [3]],
            ),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = pd.DataFrame(
        [
            [-4, -4, "category1", "a", 1],
            [-3, -3, "category1", "a", 1],
            [-2, -2, "category1", "a", 1],
            [-1, -1, "category1", "b", 1],
            [0, 0, "category1", "b", 1],
            [1, 1, "category1", "b", 1],
            [2, 2, "category1", "c", 1],
            [3, 3, "category1", "c", 1],
            [2, 3, "category1", "c", 1],
            [3, 1, "category1", "a", 1],
            [3, 4, "category1", "a", 0],
            [4, 4, "category2", "b", 0],
            [1, 4, "category2", "b", 0],
            [1, 0, "category2", "c", 0],
            [1, 2, "category2", "c", 0],
            [2, 4, "category2", "a", 1],
        ],
        columns=["x_1", "x_2", "cat_x_3", "cat_x_4", "y"],
    )
    experiments["valid_y"] = 1

    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
    )
    model = surrogates.map(model)
    with pytest.raises(
        ValueError,
        match="The feature to be stratified needs to be a DiscreteInput, CategoricalInput, CategoricalOutput, or ContinuousOutput",
    ):
        model.cross_validate(experiments, folds=5, stratified_feature=key)



from typing import Dict, Optional

import numpy as np
import plotly.express as px
import torch

import bofire.priors.api as priors
from bofire.data_models.priors.api import AnyPrior


def plot_prior_pdf_plotly(
    prior: AnyPrior,
    lower: float,
    upper: float,
    layout_options: Optional[Dict] = None,
):
    """Plot the probability density function of the prior with plotly.

    Args:
        prior (AnyPrior): The prior that should be plotted.
        lower (float): lower bound for computing the prior pdf.
        upper (float): upper bound for computing the prior pdf.
        layout_options (Dict, optional): Layout options passed to plotly. Defaults to {}.

    Returns:
        fig, ax objects of the plot.
    """

    x = np.linspace(lower, upper, 1000)

    fig = px.line(
        x=x,
        y=np.exp(priors.map(prior).log_prob(torch.from_numpy(x)).numpy()),
    )

    if layout_options is not None:
        fig.update_layout(layout_options)

    return fig



import importlib

import pandas as pd
import pytest
import torch
from botorch.models import MixedSingleTaskGP, SingleTaskGP
from botorch.models.transforms.input import (
    ChainedInputTransform,
    InputStandardize,
    Normalize,
    OneHotToNumeric,
)
from botorch.models.transforms.outcome import Standardize
from pandas.testing import assert_frame_equal
from pydantic import ValidationError

import bofire.surrogates.api as surrogates
from bofire.benchmarks.api import Himmelblau
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum, RegressionMetricsEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    MolecularInput,
)
from bofire.data_models.kernels.api import (
    HammondDistanceKernel,
    MaternKernel,
    RBFKernel,
    ScaleKernel,
    TanimotoKernel,
)
from bofire.data_models.molfeatures.api import (
    Fingerprints,
    FingerprintsFragments,
    Fragments,
    MordredDescriptors,
)
from bofire.data_models.priors.api import (
    BOTORCH_LENGTHCALE_PRIOR,
    BOTORCH_NOISE_PRIOR,
    BOTORCH_SCALE_PRIOR,
    MBO_LENGTHCALE_PRIOR,
    MBO_NOISE_PRIOR,
    MBO_OUTPUTSCALE_PRIOR,
)
from bofire.data_models.surrogates.api import (
    MixedSingleTaskGPSurrogate,
    ScalerEnum,
    SingleTaskGPHyperconfig,
    SingleTaskGPSurrogate,
)
from bofire.data_models.surrogates.tanimoto_gp import TanimotoGPSurrogate
from bofire.data_models.surrogates.trainable import metrics2objectives
from bofire.surrogates.single_task_gp import get_scaler
from bofire.utils.torch_tools import tkwargs

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None


@pytest.mark.parametrize(
    "scaler_enum, input_preprocessing_specs, expected_scaler, expected_indices, expected_offset, expected_coefficient",
    [
        (
            ScalerEnum.NORMALIZE,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.ONE_HOT,
            },
            Normalize,
            torch.tensor([0, 1], dtype=torch.int64),
            torch.tensor([-4.0, -4.0]).to(**tkwargs),
            torch.tensor([8.0, 8.0]).to(**tkwargs),
        ),
        (
            ScalerEnum.NORMALIZE,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.DESCRIPTOR,
            },
            Normalize,
            torch.tensor([0, 1, 2], dtype=torch.int64),
            torch.tensor([-4.0, -4.0, 1.0]).to(**tkwargs),
            torch.tensor([8.0, 8.0, 5.0]).to(**tkwargs),
        ),
        (
            ScalerEnum.STANDARDIZE,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.ONE_HOT,
            },
            InputStandardize,
            torch.tensor([0, 1], dtype=torch.int64),
            None,
            None,
        ),
        (
            ScalerEnum.STANDARDIZE,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.DESCRIPTOR,
            },
            InputStandardize,
            torch.tensor([0, 1, 2], dtype=torch.int64),
            None,
            None,
        ),
        (
            ScalerEnum.IDENTITY,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.ONE_HOT,
            },
            type(None),
            None,
            None,
            None,
        ),
        (
            ScalerEnum.IDENTITY,
            {
                "x_cat": CategoricalEncodingEnum.ONE_HOT,
                "x_desc": CategoricalEncodingEnum.DESCRIPTOR,
            },
            type(None),
            None,
            None,
            None,
        ),
    ],
)
def test_get_scaler(
    scaler_enum,
    input_preprocessing_specs,
    expected_scaler,
    expected_indices,
    expected_offset,
    expected_coefficient,
):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [
            CategoricalInput(key="x_cat", categories=["mama", "papa"]),
            CategoricalDescriptorInput(
                key="x_desc",
                categories=["alpha", "beta"],
                descriptors=["oskar"],
                values=[[1], [6]],
            ),
        ]
    )
    experiments = inputs.sample(n=10)
    scaler = get_scaler(
        inputs=inputs,
        input_preprocessing_specs=input_preprocessing_specs,
        scaler=scaler_enum,
        X=experiments[inputs.get_keys()],
    )
    assert isinstance(scaler, expected_scaler)
    if expected_indices is not None:
        assert (scaler.indices == expected_indices).all()
    else:
        with pytest.raises(AttributeError):
            assert (scaler.indices == expected_indices).all()
    if expected_offset is not None:
        assert torch.allclose(scaler.offset, expected_offset)
        assert torch.allclose(scaler.coefficient, expected_coefficient)
    else:
        if scaler is None:
            with pytest.raises(AttributeError):
                assert (scaler.offset == expected_offset).all()
            with pytest.raises(AttributeError):
                assert (scaler.coefficient == expected_coefficient).all()


@pytest.mark.parametrize(
    "kernel, scaler",
    [
        (ScaleKernel(base_kernel=RBFKernel(ard=True)), ScalerEnum.NORMALIZE),
        (ScaleKernel(base_kernel=RBFKernel(ard=False)), ScalerEnum.STANDARDIZE),
        (ScaleKernel(base_kernel=RBFKernel(ard=False)), ScalerEnum.IDENTITY),
    ],
)
def test_SingleTaskGPModel(kernel, scaler):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    model = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        kernel=kernel,
        scaler=scaler,
    )
    model = surrogates.map(model)
    samples = inputs.sample(5)
    # test error on non fitted model
    with pytest.raises(ValueError):
        model.predict(samples)
    model.fit(experiments)
    # dump the model
    dump = model.dumps()
    # make predictions
    samples2 = samples.copy()
    samples2 = samples2.astype({"x_1": "object"})
    preds = model.predict(samples2)
    assert preds.shape == (5, 2)
    # check that model is composed correctly
    assert isinstance(model.model, SingleTaskGP)
    assert isinstance(model.model.outcome_transform, Standardize)
    if scaler == ScalerEnum.NORMALIZE:
        assert isinstance(model.model.input_transform, Normalize)
    elif scaler == ScalerEnum.STANDARDIZE:
        assert isinstance(model.model.input_transform, InputStandardize)
    else:
        with pytest.raises(AttributeError):
            assert model.model.input_transform is None
    assert model.is_compatibilized is False
    # reload the model from dump and check for equality in predictions
    model2 = SingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        kernel=kernel,
        scaler=scaler,
    )
    model2 = surrogates.map(model2)
    model2.loads(dump)
    preds2 = model2.predict(samples)
    assert_frame_equal(preds, preds2)


@pytest.mark.parametrize("target_metric", list(RegressionMetricsEnum))
def test_hyperconfig_domain(target_metric: RegressionMetricsEnum):
    # we test here also the abstract methods from the corresponding base class
    # should be move somewhere else when tidying up all tests
    hy = SingleTaskGPHyperconfig(target_metric=target_metric)
    assert hy.domain.inputs == hy.inputs
    assert hy.domain.outputs.get_keys() == [target_metric.name]
    assert hy.domain.outputs[0].objective == metrics2objectives[target_metric]()


def test_hyperconfig_invalid():
    with pytest.raises(
        ValueError,
        match="It is not allowed to scpecify the number of its for FactorialStrategy",
    ):
        SingleTaskGPHyperconfig(n_iterations=5)
    with pytest.raises(
        ValueError,
        match="At least number of hyperparams plus 2 iterations has to be specified",
    ):
        SingleTaskGPHyperconfig(n_iterations=3, hyperstrategy="RandomStrategy")
    hy = SingleTaskGPHyperconfig(n_iterations=None, hyperstrategy="RandomStrategy")
    assert hy.n_iterations == 13


def test_SingleTaskGPHyperconfig():
    # we test here also the basic trainable
    benchmark = Himmelblau()
    surrogate_data_no_hy = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs,
        outputs=benchmark.domain.outputs,
        hyperconfig=None,
    )
    with pytest.raises(ValueError, match="No hyperconfig available."):
        surrogate_data_no_hy.update_hyperparameters(
            benchmark.domain.inputs.sample(1).loc[0]
        )
    # test that correct stuff is written
    surrogate_data = SingleTaskGPSurrogate(
        inputs=benchmark.domain.inputs, outputs=benchmark.domain.outputs
    )
    candidate = surrogate_data.hyperconfig.inputs.sample(1).loc[0]
    surrogate_data.update_hyperparameters(candidate)
    assert surrogate_data.kernel.base_kernel.ard == (candidate["ard"] == "True")
    if candidate.kernel == "matern_1.5":
        assert isinstance(surrogate_data.kernel.base_kernel, MaternKernel)
        assert surrogate_data.kernel.base_kernel.nu == 1.5
    elif candidate.kernel == "matern_2.5":
        assert isinstance(surrogate_data.kernel.base_kernel, MaternKernel)
        assert surrogate_data.kernel.base_kernel.nu == 2.5
    else:
        assert isinstance(surrogate_data.kernel.base_kernel, RBFKernel)
    if candidate.prior == "mbo":
        assert surrogate_data.noise_prior == MBO_NOISE_PRIOR()
        assert surrogate_data.kernel.outputscale_prior == MBO_OUTPUTSCALE_PRIOR()
        assert (
            surrogate_data.kernel.base_kernel.lengthscale_prior
            == MBO_LENGTHCALE_PRIOR()
        )
    else:
        assert surrogate_data.noise_prior == BOTORCH_NOISE_PRIOR()
        assert surrogate_data.kernel.outputscale_prior == BOTORCH_SCALE_PRIOR()
        assert (
            surrogate_data.kernel.base_kernel.lengthscale_prior
            == BOTORCH_LENGTHCALE_PRIOR()
        )


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
@pytest.mark.parametrize(
    "kernel, specs",
    [
        (
            ScaleKernel(base_kernel=TanimotoKernel(ard=False)),
            {"x_1": Fingerprints(n_bits=32)},
        ),
        (
            ScaleKernel(base_kernel=TanimotoKernel(ard=True)),
            {"x_1": Fragments()},
        ),
        (
            ScaleKernel(base_kernel=TanimotoKernel(ard=False)),
            {"x_1": FingerprintsFragments(n_bits=32)},
        ),
        (
            ScaleKernel(base_kernel=RBFKernel(ard=True)),
            {"x_1": MordredDescriptors(descriptors=["NssCH2", "ATSC2d"])},
        ),
    ],
)
def test_TanimotoGP(kernel, specs):
    inputs = Inputs(features=[MolecularInput(key="x_1")])
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = [
        ["CC(=O)Oc1ccccc1C(=O)O", 88.0],
        ["c1ccccc1", 35.0],
        ["[CH3][CH2][OH]", 69.0],
        ["N[C@](C)(F)C(=O)O", 20.0],
    ]
    experiments = pd.DataFrame(experiments, columns=["x_1", "y"])
    experiments["valid_y"] = 1
    model = TanimotoGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        kernel=kernel,
        input_preprocessing_specs=specs,
    )
    model = surrogates.map(model)
    model.fit(experiments)
    # dump the model
    dump = model.dumps()
    # make predictions
    preds = model.predict(experiments.iloc[:-1])
    assert preds.shape == (3, 2)
    # check that model is composed correctly
    assert isinstance(model.model, SingleTaskGP)
    assert isinstance(model.model.outcome_transform, Standardize)
    assert model.is_compatibilized is False
    # reload the model from dump and check for equality in predictions
    model2 = TanimotoGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        kernel=kernel,
        input_preprocessing_specs=specs,
    )
    model2 = surrogates.map(model2)
    model2.loads(dump)
    preds2 = model2.predict(experiments.iloc[:-1])
    assert_frame_equal(preds, preds2)


def test_MixedGPModel_invalid_preprocessing():
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments["valid_y"] = 1
    with pytest.raises(ValidationError):
        MixedSingleTaskGPSurrogate(
            inputs=inputs,
            outputs=outputs,
        )


@pytest.mark.parametrize(
    "kernel, scaler",
    [
        (RBFKernel(ard=True), ScalerEnum.NORMALIZE),
        (RBFKernel(ard=False), ScalerEnum.STANDARDIZE),
        (RBFKernel(ard=False), ScalerEnum.IDENTITY),
    ],
)
def test_MixedGPModel(kernel, scaler):
    inputs = Inputs(
        features=[
            ContinuousInput(
                key=f"x_{i+1}",
                bounds=(-4, 4),
            )
            for i in range(2)
        ]
        + [CategoricalInput(key="x_cat", categories=["mama", "papa"])]
    )
    outputs = Outputs(features=[ContinuousOutput(key="y")])
    experiments = inputs.sample(n=10)
    experiments.eval("y=((x_1**2 + x_2 - 11)**2+(x_1 + x_2**2 -7)**2)", inplace=True)
    experiments.loc[experiments.x_cat == "mama", "y"] *= 5.0
    experiments.loc[experiments.x_cat == "papa", "y"] /= 2.0
    experiments["valid_y"] = 1

    model = MixedSingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        input_preprocessing_specs={"x_cat": CategoricalEncodingEnum.ONE_HOT},
        scaler=scaler,
        continuous_kernel=kernel,
        categorical_kernel=HammondDistanceKernel(),
    )
    model = surrogates.map(model)
    with pytest.raises(ValueError):
        model.dumps()
    model.fit(experiments)
    # dump the model
    dump = model.dumps()
    # make predictions
    samples = inputs.sample(5)
    preds = model.predict(samples)
    assert preds.shape == (5, 2)
    # check that model is composed correctly
    assert isinstance(model.model, MixedSingleTaskGP)
    assert isinstance(model.model.outcome_transform, Standardize)
    if scaler == ScalerEnum.NORMALIZE:
        assert isinstance(model.model.input_transform, ChainedInputTransform)
        assert isinstance(model.model.input_transform.tf1, Normalize)
        assert torch.eq(
            model.model.input_transform.tf1.indices,
            torch.tensor([0, 1], dtype=torch.int64),
        ).all()
        assert isinstance(model.model.input_transform.tf2, OneHotToNumeric)
    elif scaler == ScalerEnum.STANDARDIZE:
        assert isinstance(model.model.input_transform, ChainedInputTransform)
        assert isinstance(model.model.input_transform.tf1, InputStandardize)
        assert torch.eq(
            model.model.input_transform.tf1.indices,
            torch.tensor([0, 1], dtype=torch.int64),
        ).all()
        assert isinstance(model.model.input_transform.tf2, OneHotToNumeric)
    else:
        assert isinstance(model.model.input_transform, OneHotToNumeric)
    assert model.is_compatibilized is False
    # reload the model from dump and check for equality in predictions
    model2 = MixedSingleTaskGPSurrogate(
        inputs=inputs,
        outputs=outputs,
        continuous_kernel=kernel,
        scaler=scaler,
    )
    model2 = surrogates.map(model2)
    model2.loads(dump)
    preds2 = model2.predict(samples)
    assert_frame_equal(preds, preds2)



from abc import abstractmethod
from typing import List, Optional, Tuple

import numpy as np
import pandas as pd
from pydantic import PositiveInt

from bofire.data_models.features.api import TInputTransformSpecs
from bofire.data_models.strategies.api import Strategy as DataModel
from bofire.strategies.data_models.candidate import Candidate
from bofire.strategies.data_models.values import InputValue, OutputValue
from bofire.strategies.strategy import Strategy


class PredictiveStrategy(Strategy):
    """Base class for all model based strategies.

    Provides abstract scaffold for fit, predict, and calc_acquistion methods.
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.is_fitted = False

    @property
    @abstractmethod
    def input_preprocessing_specs(self) -> TInputTransformSpecs:
        pass

    def ask(
        self,
        candidate_count: Optional[PositiveInt] = None,
        add_pending: bool = False,
        raise_validation_error: bool = True,
    ) -> pd.DataFrame:
        """Function to generate new candidates.

        Args:
            candidate_count (PositiveInt, optional): Number of candidates to be generated. If not provided, the number of candidates is determined automatically. Defaults to None.
            add_pending (bool, optional): If true the proposed candidates are added to the set of pending experiments. Defaults to False.
            raise_validation_error (bool, optional): If true an error will be raised if candidates violate constraints,
                otherwise only a warning will be displayed. Defaults to True.

        Returns:
            pd.DataFrame: DataFrame with candidates (proposed experiments)
        """
        candidates = super().ask(
            candidate_count=candidate_count,
            add_pending=add_pending,
            raise_validation_error=raise_validation_error,
        )
        self.domain.validate_candidates(
            candidates=candidates, raise_validation_error=raise_validation_error
        )
        return candidates

    def tell(
        self,
        experiments: pd.DataFrame,
        replace: bool = False,
        retrain: bool = True,
    ):
        """This function passes new experimental data to the optimizer.

        Args:
            experiments (pd.DataFrame): DataFrame with experimental data
            replace (bool, optional): Boolean to decide if the experimental data should replace the former dataFrame or if the new experiments should be attached. Defaults to False.
            retrain (bool, optional): If True, model(s) are retrained when new experimental data is passed to the optimizer. Defaults to True.
        """
        # maybe unite the preprocessor here with the one of the parent tell
        # TODO: add self.domain.validate_experiments(self.experiments, strict=True) here to ensure variance in each feature?
        if len(experiments) == 0:
            return
        if replace:
            self.set_experiments(experiments)
        else:
            self.add_experiments(experiments)
        if retrain and self.has_sufficient_experiments():
            self.fit()
            # we have a seperate _tell here for things that are relevant when setting up the strategy but unrelated
            # to fitting the models like initializing the ACQF.
            self._tell()

    def predict(self, experiments: pd.DataFrame) -> pd.DataFrame:
        """Run predictions for the provided experiments. Only input features have to be provided.

        Args:
            experiments (pd.DataFrame): Experimental data for which predictions should be performed.

        Returns:
            pd.DataFrame: Dataframe with the predicted values.
        """
        if self.is_fitted is not True:
            raise ValueError("Model not yet fitted.")
        # TODO: validate also here the experiments but only for the input_columns
        # transformed = self.transformer.transform(experiments)
        transformed = self.domain.inputs.transform(
            experiments=experiments, specs=self.input_preprocessing_specs
        )
        preds, stds = self._predict(transformed)
        if stds is not None:
            predictions = pd.DataFrame(
                data=np.hstack((preds, stds)),
                columns=["%s_pred" % feat.key for feat in self.domain.outputs.get()]
                + ["%s_sd" % feat.key for feat in self.domain.outputs.get()],
            )
        else:
            predictions = pd.DataFrame(
                data=preds,
                columns=["%s_pred" % feat.key for feat in self.domain.outputs.get()],
            )
        desis = self.domain.outputs(predictions, predictions=True)
        predictions = pd.concat((predictions, desis), axis=1)
        predictions.index = experiments.index
        return predictions

    @abstractmethod
    def _predict(self, experiments: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Abstract method in which the actual prediction is happening. Has to be overwritten."""
        pass

    def fit(self):
        """Fit the model(s) to the experimental data."""
        assert (
            self.experiments is not None and len(self.experiments) > 0
        ), "No fitting data available"
        self.domain.validate_experiments(self.experiments, strict=True)
        # transformed = self.transformer.fit_transform(self.experiments)
        self._fit(self.experiments)
        self.is_fitted = True

    @abstractmethod
    def _fit(self, experiments: pd.DataFrame):
        """Abstract method where the acutal prediction are occuring."""
        pass

    def to_candidates(self, candidates: pd.DataFrame) -> List[Candidate]:
        """Transform candiadtes dataframe to a list of `Candidate` objects.

        Args:
            candidates (pd.DataFrame): candidates formatted as dataframe

        Returns:
            List[Candidate]: candidates formatted as list of `Candidate` objects.
        """
        return [
            Candidate(
                inputValues={
                    key: InputValue(value=row[key])
                    for key in self.domain.inputs.get_keys()
                },
                outputValues={
                    feat.key: OutputValue(
                        predictedValue=row[f"{feat.key}_pred"],
                        standardDeviation=row[f"{feat.key}_sd"],
                        objective=row[f"{feat.key}_des"]
                        if feat.objective is not None  # type: ignore
                        else 1.0,
                    )
                    for feat in self.domain.outputs.get()
                },
            )
            for _, row in candidates.iterrows()
        ]



from typing import Literal, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import root_validator

from bofire.data_models.objectives.objective import Objective, TWeight


class IdentityObjective(Objective):
    """An objective returning the identity as reward.
    The return can be scaled, when a lower and upper bound are provided.

    Attributes:
        w (float): float between zero and one for weighting the objective
        bounds (Tuple[float], optional): Bound for normalizing the objective between zero and one. Defaults to (0,1).
    """

    type: Literal["IdentityObjective"] = "IdentityObjective"
    w: TWeight = 1
    bounds: Tuple[float, float] = (0, 1)

    @property
    def lower_bound(self) -> float:
        return self.bounds[0]

    @property
    def upper_bound(self) -> float:
        return self.bounds[1]

    @root_validator(pre=False, skip_on_failure=True)
    def validate_lower_upper(cls, values):
        """Validation function to ensure that lower bound is always greater the upper bound

        Args:
            values (Dict): The attributes of the class

        Raises:
            ValueError: when a lower bound higher than the upper bound is passed

        Returns:
            Dict: The attributes of the class
        """
        if values["bounds"][0] > values["bounds"][1]:
            raise ValueError(
                f'lower bound must be <= upper bound, got {values["lower_bound"]} > {values["upper_bound"]}'
            )
        return values

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """The call function returning a reward for passed x values

        Args:
            x (np.ndarray): An array of x values

        Returns:
            np.ndarray: The identity as reward, might be normalized to the passed lower and upper bounds
        """
        return (x - self.lower_bound) / (self.upper_bound - self.lower_bound)


class MaximizeObjective(IdentityObjective):
    """Child class from the identity function without modifications, since the parent class is already defined as maximization

    Attributes:
        w (float): float between zero and one for weighting the objective
        bounds (Tuple[float], optional): Bound for normalizing the objective between zero and one. Defaults to (0,1).
    """

    type: Literal["MaximizeObjective"] = "MaximizeObjective"


class MinimizeObjective(IdentityObjective):
    """Class returning the negative identity as reward.

    Attributes:
        w (float): float between zero and one for weighting the objective
        bounds (Tuple[float], optional): Bound for normalizing the objective between zero and one. Defaults to (0,1).
    """

    type: Literal["MinimizeObjective"] = "MinimizeObjective"

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """The call function returning a reward for passed x values

        Args:
            x (np.ndarray): An array of x values

        Returns:
            np.ndarray: The negative identity as reward, might be normalized to the passed lower and upper bounds
        """
        return -1.0 * (x - self.lower_bound) / (self.upper_bound - self.lower_bound)



from typing import Dict, Optional

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from bofire.data_models.features.api import ContinuousOutput


def plot_objective_plotly(
    feature: ContinuousOutput,
    lower: float,
    upper: float,
    values: Optional[pd.Series] = None,
    layout_options: Optional[Dict] = None,
):
    """Plot the assigned objective.

    Args:
        feature (ContinuousOutput): Output feature, whose objective should be visualized.
        lower (float): lower bound for the plot
        upper (float): upper bound for the plot
        values (Optional[pd.Series], optional): If provided, scatter also the historical data in the plot. Defaults to None.
        layout_options: (Dict, optional): Options that are passed to plotlys `update_layout`.
    """
    if feature.objective is None:
        raise ValueError(
            f"No objective assigned for ContinuousOutputFeauture with key {feature.key}."
        )

    x = pd.Series(np.linspace(lower, upper, 5000))
    reward = feature.objective.__call__(x)

    fig1 = px.line(x=x, y=reward, title=feature.key)

    if values is not None:
        fig2 = px.scatter(
            x=values,
            y=feature.objective.__call__(values),
        )
        fig = go.Figure(data=fig1.data + fig2.data)  # type: ignore
    else:
        fig = fig1

    if layout_options is not None:
        fig.update_layout(layout_options)

    return fig



import random

import bofire.data_models.priors.api as priors
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    priors.NormalPrior,
    lambda: {
        "loc": random.random(),
        "scale": random.random(),
    },
)
specs.add_valid(
    priors.GammaPrior,
    lambda: {
        "concentration": random.random(),
        "rate": random.random(),
    },
)



from typing import Union

from bofire.data_models.constraints.constraint import (
    Constraint,
    ConstraintError,
    ConstraintNotFulfilledError,
)
from bofire.data_models.constraints.linear import (
    LinearConstraint,
    LinearEqualityConstraint,
    LinearInequalityConstraint,
)
from bofire.data_models.constraints.nchoosek import NChooseKConstraint
from bofire.data_models.constraints.nonlinear import (
    NonlinearConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
)

AbstractConstraint = Union[
    Constraint,
    LinearConstraint,
    NonlinearConstraint,
]

AnyConstraint = Union[
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NonlinearEqualityConstraint,
    NonlinearInequalityConstraint,
    NChooseKConstraint,
]

AnyConstraintError = Union[ConstraintError, ConstraintNotFulfilledError]

# %%



from typing import Dict, Tuple, Type

import pandas as pd

import bofire.data_models.strategies.api as data_models
import bofire.strategies.stepwise.conditions as conditions
from bofire.data_models.strategies.api import Step
from bofire.data_models.strategies.api import StepwiseStrategy as data_model
from bofire.strategies.doe_strategy import DoEStrategy  # noqa: F401
from bofire.strategies.predictives.qehvi import QehviStrategy  # noqa: F401
from bofire.strategies.predictives.qnehvi import QnehviStrategy  # noqa: F401
from bofire.strategies.predictives.qparego import QparegoStrategy  # noqa: F401
from bofire.strategies.predictives.sobo import (  # noqa: F401
    AdditiveSoboStrategy,
    MultiplicativeSoboStrategy,
    SoboStrategy,
)
from bofire.strategies.random import RandomStrategy  # noqa: F401
from bofire.strategies.samplers.polytope import PolytopeSampler  # noqa: F401
from bofire.strategies.samplers.rejection import RejectionSampler  # noqa: F401
from bofire.strategies.strategy import Strategy  # noqa: F401

# we have to duplicate the map functionality due to prevent circular imports
STRATEGY_MAP: Dict[Type[data_models.Strategy], Type[Strategy]] = {
    data_models.RandomStrategy: RandomStrategy,
    data_models.SoboStrategy: SoboStrategy,
    data_models.AdditiveSoboStrategy: AdditiveSoboStrategy,
    data_models.MultiplicativeSoboStrategy: MultiplicativeSoboStrategy,
    data_models.QehviStrategy: QehviStrategy,
    data_models.QnehviStrategy: QnehviStrategy,
    data_models.QparegoStrategy: QparegoStrategy,
    data_models.PolytopeSampler: PolytopeSampler,
    data_models.RejectionSampler: RejectionSampler,
    data_models.DoEStrategy: DoEStrategy,
}


def map(data_model: data_models.Strategy) -> Strategy:
    cls = STRATEGY_MAP[data_model.__class__]
    return cls.from_spec(data_model=data_model)


class StepwiseStrategy(Strategy):
    def __init__(self, data_model: data_model, **kwargs):
        super().__init__(data_model, **kwargs)
        self.steps = data_model.steps

    def has_sufficient_experiments(self) -> bool:
        return True

    def _get_step(self) -> Tuple[int, Step]:  # type: ignore
        for i, step in enumerate(self.steps):
            condition = conditions.map(step.condition)
            if condition.evaluate(self.domain, experiments=self.experiments):
                return i, step
        raise ValueError("No condition could be satisfied.")

    def _ask(self, candidate_count: int) -> pd.DataFrame:
        # we have to decide here w
        istep, step = self._get_step()
        if (step.max_parallelism > 0) and (candidate_count > step.max_parallelism):
            raise ValueError(
                f"Maximum number of candidates for step {istep} is {step.max_parallelism}."
            )
        # map it
        strategy = map(step.strategy_data)
        # tell the experiments
        if self.num_experiments > 0:
            strategy.tell(experiments=self.experiments)
        # tell pending
        if self.num_candidates > 0:
            strategy.set_candidates(self.candidates)
        # ask and return
        return strategy.ask(candidate_count=candidate_count)



from typing import Literal

from pydantic import conint, validator

from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class SaasSingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    type: Literal["SaasSingleTaskGPSurrogate"] = "SaasSingleTaskGPSurrogate"
    warmup_steps: conint(ge=1) = 256  # type: ignore
    num_samples: conint(ge=1) = 128  # type: ignore
    thinning: conint(ge=1) = 16  # type: ignore
    scaler: ScalerEnum = ScalerEnum.NORMALIZE

    @validator("thinning")
    def validate_thinning(cls, value, values):
        if values["num_samples"] / value < 1:
            raise ValueError("`num_samples` has to be larger than `thinning`.")
        return value



import numpy as np
import pandas as pd
from formulaic import Formula

from bofire.data_models.domain.api import Domain
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.strategies.doe.objective import (
    AOptimality,
    DOptimality,
    EOptimality,
    GOptimality,
    Objective,
    SpaceFilling,
)
from bofire.strategies.doe.utils import get_formula_from_string


def test_Objective_model_jacobian_t():
    # "small" model
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    vars = domain.inputs.get_keys()
    f = Formula("x1 + x2 + x3 + x1:x2 + {x3**2}")
    x = np.array([[1, 2, 3]])

    objective = Objective(
        domain=domain,
        model=f,
        n_experiments=1,
    )
    model_jacobian_t = objective._model_jacobian_t

    B = np.zeros(shape=(3, 6))
    B[:, 1:4] = np.eye(3)
    B[:, 4] = np.array([0, 0, 6])
    B[:, 5] = np.array([2, 1, 0])

    assert np.allclose(B, model_jacobian_t(x))

    # fully quadratic model
    f = Formula("x1 + x2 + x3 + x1:x2 + x1:x3 + x2:x3 + {x1**2} + {x2**2} + {x3**2}")
    model_terms = np.array(f, dtype=str)
    x = np.array([[1, 2, 3]])

    objective = Objective(
        domain=domain,
        model=f,
        n_experiments=1,
    )
    model_jacobian_t = objective._model_jacobian_t
    B = np.zeros(shape=(3, 10))
    B[:, 1:4] = np.eye(3)
    B[:, 4:7] = 2 * np.diag(x[0])
    B[:, 7:] = np.array([[2, 1, 0], [3, 0, 1], [0, 3, 2]]).T
    B = pd.DataFrame(
        B,
        columns=[
            "1",
            "x1",
            "x2",
            "x3",
            "x1**2",
            "x2**2",
            "x3**2",
            "x1:x2",
            "x1:x3",
            "x2:x3",
        ],
    )
    B = B[model_terms].to_numpy()

    assert np.allclose(B, model_jacobian_t(x)[0])

    # fully cubic model
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(5)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    vars = ["x1", "x2", "x3", "x4", "x5"]
    n_vars = len(vars)

    formula = ""
    for name in vars:
        formula += name + " + "

    for name in vars:
        formula += "{" + name + "**2} + "
    for i in range(n_vars):
        for j in range(i + 1, n_vars):
            term = str(Formula(vars[j] + ":" + vars[i] + "-1")) + " + "
            formula += term

    for name in vars:
        formula += "{" + name + "**3} + "
    for i in range(n_vars):
        for j in range(i + 1, n_vars):
            for k in range(j + 1, n_vars):
                term = (
                    str(Formula(vars[k] + ":" + vars[j] + ":" + vars[i] + "-1")) + " + "
                )
                formula += term
    f = Formula(formula[:-3])
    x = np.array([[1, 2, 3, 4, 5]])
    objective = Objective(
        domain=domain,
        model=f,
        n_experiments=1,
    )
    model_jacobian_t = objective._model_jacobian_t

    B = np.array(
        [
            [
                0.0,
                1.0,
                2.0,
                3.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                2.0,
                3.0,
                4.0,
                5.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                6.0,
                8.0,
                10.0,
                12.0,
                15.0,
                20.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
            [
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                4.0,
                12.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                3.0,
                4.0,
                5.0,
                0.0,
                0.0,
                0.0,
                3.0,
                4.0,
                5.0,
                0.0,
                0.0,
                0.0,
                12.0,
                15.0,
                20.0,
                0.0,
            ],
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                6.0,
                27.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                2.0,
                0.0,
                0.0,
                4.0,
                5.0,
                0.0,
                2.0,
                0.0,
                0.0,
                4.0,
                5.0,
                0.0,
                8.0,
                10.0,
                0.0,
                20.0,
            ],
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                8.0,
                48.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                2.0,
                0.0,
                3.0,
                0.0,
                5.0,
                0.0,
                2.0,
                0.0,
                3.0,
                0.0,
                5.0,
                6.0,
                0.0,
                10.0,
                15.0,
            ],
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                10.0,
                75.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                2.0,
                0.0,
                3.0,
                4.0,
                0.0,
                0.0,
                2.0,
                0.0,
                3.0,
                4.0,
                0.0,
                6.0,
                8.0,
                12.0,
            ],
        ]
    )

    B = pd.DataFrame(
        B,
        columns=[
            "1",
            "x1",
            "x1**2",
            "x1**3",
            "x2",
            "x2**2",
            "x2**3",
            "x3",
            "x3**2",
            "x3**3",
            "x4",
            "x4**2",
            "x4**3",
            "x5",
            "x5**2",
            "x5**3",
            "x2:x1",
            "x3:x1",
            "x4:x1",
            "x5:x1",
            "x3:x2",
            "x4:x2",
            "x5:x2",
            "x4:x3",
            "x5:x3",
            "x5:x4",
            "x3:x2:x1",
            "x4:x2:x1",
            "x5:x2:x1",
            "x4:x3:x1",
            "x5:x3:x1",
            "x5:x4:x1",
            "x4:x3:x2",
            "x5:x3:x2",
            "x5:x4:x2",
            "x5:x4:x3",
        ],
    )

    B = B[objective.model_terms].to_numpy()

    assert np.allclose(B, model_jacobian_t(x)[0])


def test_Objective_convert_input_to_model_tensor():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    d_optimality = DOptimality(domain=domain, model=model, n_experiments=3)
    x = np.array([1, 0, 0, 0, 2, 0, 0, 0, 3])
    print(domain.inputs)
    X = d_optimality._convert_input_to_model_tensor(x).detach().numpy()
    assert np.allclose(X, np.array([[1, 1, 0, 0], [1, 0, 2, 0], [1, 0, 0, 3]]))


def test_DOptimality_instantiation():
    # default jacobian building block
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    model = Formula("x1 + x2 + x3 + x1:x2 + {x3**2}")

    d_optimality = DOptimality(
        domain=domain,
        model=model,
        n_experiments=2,
    )

    assert isinstance(d_optimality.domain, Domain)
    assert all(
        np.array(d_optimality.domain.inputs.get_keys()) == np.array(["x1", "x2", "x3"])
    )
    for i in d_optimality.domain.inputs.get():
        assert isinstance(i, ContinuousInput)
        assert i.upper_bound == 1
        assert i.lower_bound == 0
    assert all(np.array(d_optimality.domain.outputs.get_keys()) == np.array(["y"]))

    assert isinstance(d_optimality.model, Formula)
    assert all(
        np.array(d_optimality.model, dtype=str)
        == np.array(["1", "x1", "x2", "x3", "x3**2", "x1:x2"])
    )

    x = np.array([[1, 2, 3], [1, 2, 3]])
    B = np.zeros(shape=(3, 6))
    B[:, 1:4] = np.eye(3)
    B[:, 4] = np.array([0, 0, 6])
    B[:, 5] = np.array([2, 1, 0])

    assert np.allclose(B, d_optimality._model_jacobian_t(x))
    assert np.shape(
        d_optimality.evaluate_jacobian(np.array([[1, 1, 1], [2, 2, 2]]).flatten())
    ) == (6,)

    # 5th order model
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    model = Formula("{x1**5} + {x2**5} + {x3**5}")

    d_optimality = DOptimality(
        domain=domain,
        model=model,
        n_experiments=3,
    )

    x = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
    B = np.zeros(shape=(3, 4))
    B[:, 1:] = 5 * np.diag(x[0] ** 4)

    assert np.allclose(B, d_optimality._model_jacobian_t(x))
    assert np.shape(
        d_optimality.evaluate_jacobian(
            np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]]).flatten()
        )
    ) == (9,)


def test_DOptimality_evaluate_jacobian():
    # n_experiment = 1, n_inputs = 2, model: x1 + x2
    def jacobian(x: np.ndarray, delta=1e-3) -> np.ndarray:  # type: ignore
        return -2 * x / (x[0] ** 2 + x[1] ** 2 + delta)

    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(2)
        ],
        outputs=[ContinuousOutput(key="y")],
    )

    model = Formula("x1 + x2 - 1")
    n_experiments = 1
    d_optimality = DOptimality(
        domain=domain, model=model, n_experiments=n_experiments, delta=1e-3
    )

    np.random.seed(1)
    for _ in range(10):
        x = np.random.rand(2)
        assert np.allclose(d_optimality.evaluate_jacobian(x), jacobian(x), rtol=1e-3)

    # n_experiment = 1, n_inputs = 2, model: x1**2 + x2**2
    def jacobian(x: np.ndarray, delta=1e-3) -> np.ndarray:  # type: ignore
        return -4 * x**3 / (x[0] ** 4 + x[1] ** 4 + delta)

    model = Formula("{x1**2} + {x2**2} - 1")
    d_optimality = DOptimality(
        domain=domain, model=model, n_experiments=n_experiments, delta=1e-3
    )
    np.random.seed(1)
    for _ in range(10):
        x = np.random.rand(2)
        assert np.allclose(d_optimality.evaluate_jacobian(x), jacobian(x), rtol=1e-3)

    # n_experiment = 2, n_inputs = 2, model = x1 + x2
    def jacobian(x: np.ndarray, delta=1e-3) -> np.ndarray:  # type: ignore
        X = x.reshape(2, 2)

        y = np.empty(4)
        y[0] = (
            -2
            * (
                x[0] * (x[1] ** 2 + x[3] ** 2 + delta)
                - x[1] * (x[0] * x[1] + x[2] * x[3])
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[1] = (
            -2
            * (
                x[1] * (x[0] ** 2 + x[2] ** 2 + delta)
                - x[0] * (x[0] * x[1] + x[2] * x[3])
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[2] = (
            -2
            * (
                x[2] * (x[1] ** 2 + x[3] ** 2 + delta)
                - x[3] * (x[0] * x[1] + x[2] * x[3])
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[3] = (
            -2
            * (
                x[3] * (x[0] ** 2 + x[2] ** 2 + delta)
                - x[2] * (x[0] * x[1] + x[2] * x[3])
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )

        return y

    model = Formula("x1 + x2 - 1")
    n_experiments = 2
    d_optimality = DOptimality(
        domain=domain, model=model, n_experiments=n_experiments, delta=1e-3
    )
    np.random.seed(1)
    for _ in range(10):
        x = np.random.rand(4)
        assert np.allclose(d_optimality.evaluate_jacobian(x), jacobian(x), rtol=1e-3)

    # n_experiment = 2, n_inputs = 2, model = x1**2 + x2**2
    def jacobian(x: np.ndarray, delta=1e-3) -> np.ndarray:
        X = x.reshape(2, 2) ** 2

        y = np.empty(4)
        y[0] = (
            -4
            * (
                x[0] ** 3 * (x[1] ** 4 + x[3] ** 4 + delta)
                - x[0] * x[1] ** 2 * (x[0] ** 2 * x[1] ** 2 + x[2] ** 2 * x[3] ** 2)
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[1] = (
            -4
            * (
                x[1] ** 3 * (x[0] ** 4 + x[2] ** 4 + delta)
                - x[1] * x[0] ** 2 * (x[0] ** 2 * x[1] ** 2 + x[2] ** 2 * x[3] ** 2)
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[2] = (
            -4
            * (
                x[2] ** 3 * (x[1] ** 4 + x[3] ** 4 + delta)
                - x[2] * x[3] ** 2 * (x[0] ** 2 * x[1] ** 2 + x[2] ** 2 * x[3] ** 2)
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )
        y[3] = (
            -4
            * (
                x[3] ** 3 * (x[0] ** 4 + x[2] ** 4 + delta)
                - x[3] * x[2] ** 2 * (x[0] ** 2 * x[1] ** 2 + x[2] ** 2 * x[3] ** 2)
            )
            / np.linalg.det(X.T @ X + delta * np.eye(2))
        )

        return y

    model = Formula("{x1**2} + {x2**2} - 1")
    d_optimality = DOptimality(
        domain=domain, model=model, n_experiments=n_experiments, delta=1e-3
    )

    np.random.seed(1)
    for _ in range(10):
        x = np.random.rand(4)
        assert np.allclose(d_optimality.evaluate_jacobian(x), jacobian(x), rtol=1e-3)


def test_DOptimality_evaluate():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    d_optimality = DOptimality(domain=domain, model=model, n_experiments=3)
    x = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1])
    assert np.allclose(d_optimality.evaluate(x), -np.log(4) - np.log(1e-7))


def test_AOptimality_evaluate():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key=f"x{i+1}",
                bounds=(0, 1),
            )
            for i in range(3)
        ],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    a_optimality = AOptimality(domain=domain, model=model, n_experiments=4)

    x = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0])
    assert np.allclose(a_optimality.evaluate(x), 3 * 1.9999991 + 0.9999996)


def test_AOptimality_evaluate_jacobian():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    a_optimality = AOptimality(domain=domain, model=model, n_experiments=2, delta=0)

    x = np.array([1, 0.5])

    def grad(x):
        return (
            2
            * np.array([-(x[0] * x[1] + x[1] ** 2 + 2), (x[0] * x[1] + x[0] ** 2 + 2)])
            / (x[0] - x[1]) ** 3
        )

    assert np.allclose(a_optimality.evaluate_jacobian(x), grad(x))


def test_EOptimality_evaluate():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    e_optimality = EOptimality(domain=domain, model=model, n_experiments=2, delta=0)

    x = np.array([1, 0.5])

    min_eigval = 0.5 * (
        x[0] ** 2
        - np.sqrt(
            x[0] ** 4 + 2 * x[0] ** 2 * x[1] ** 2 + 8 * x[0] * x[1] + x[1] ** 4 + 4
        )
        + x[1] ** 2
        + 2
    )

    assert np.allclose(e_optimality.evaluate(x), -min_eigval)


def test_EOptimality_evaluate_jacobian():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    e_optimality = EOptimality(domain=domain, model=model, n_experiments=2, delta=0)

    x = np.array([1, 0.5])

    def grad(x):
        temp = np.sqrt(
            x[0] ** 4 + 2 * x[0] ** 2 * x[1] ** 2 + 8 * x[0] * x[1] + x[1] ** 4 + 4
        )
        return np.array(
            [
                (x[0] ** 3 + x[0] * x[1] ** 2 + 2 * x[1]) / temp - x[0],
                (x[1] ** 3 + x[1] * x[0] ** 2 + 2 * x[0]) / temp - x[1],
            ]
        )

    assert np.allclose(e_optimality.evaluate_jacobian(x), grad(x))


def test_GOptimality_evaluate():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    g_optimality = GOptimality(domain=domain, model=model, n_experiments=2, delta=0)

    x = np.array([1, 0.5])

    # all eigenvalues are 1 since A = [[1,1],[1,0.5]] is invertible and therefore H = A (A.T A)^-1 A.T = A A^-1 A.T^-1 A.T = 1

    assert np.allclose(g_optimality.evaluate(x), 1)


def test_GOptimality_evaluate_jacobian():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    g_optimality = GOptimality(domain=domain, model=model, n_experiments=2, delta=0)

    x = np.array([1, 0.5])

    # all eigenvalues are 1 since A = [[1,1],[1,0.5]] is invertible and therefore H = A (A.T A)^-1 A.T = A A^-1 A.T^-1 A.T = 1
    # thus, the jacobian vanishes.

    assert np.allclose(g_optimality.evaluate_jacobian(x), np.zeros(2))


def test_SpaceFilling_evaluate():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    space_filling = SpaceFilling(domain=domain, model=model, n_experiments=4, delta=0)

    x = np.array([1, 0.6, 0.1, 0.3])

    assert np.allclose(space_filling.evaluate(x), -1.4)


def test_SpaceFilling_evaluate_jacobian():
    domain = Domain.from_lists(
        inputs=[ContinuousInput(key="x1", bounds=(0, 1))],
        outputs=[ContinuousOutput(key="y")],
    )
    model = get_formula_from_string("linear", domain=domain)

    space_filling = SpaceFilling(domain=domain, model=model, n_experiments=4, delta=0)

    x = np.array([1, 0.4, 0, 0.1])

    assert np.allclose(space_filling.evaluate_jacobian(x), [-1, -1, 2, 0])



# type: ignore
# this is based on the mlflow implementation: https://github.com/mlflow/mlflow/blob/master/mlflow/pytorch/pickle_module.py
import warnings

try:
    from pickle import Unpickler  # noqa: F401

    from cloudpickle import *  # noqa: F401, F403,
    from cloudpickle import CloudPickler as Pickler  # noqa: F401

except ModuleNotFoundError:
    warnings.warn("Cloudpickle is not available.")



import codecs
import pickle
from typing import Optional

import numpy as np
import pandas as pd
import torch
from botorch.models.ensemble import EnsembleModel
from sklearn.ensemble import RandomForestRegressor
from sklearn.utils.validation import check_is_fitted
from torch import Tensor

from bofire.data_models.enum import OutputFilteringEnum
from bofire.data_models.surrogates.api import RandomForestSurrogate as DataModel
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.utils.torch_tools import tkwargs


class _RandomForest(EnsembleModel):
    """Botorch wrapper around the sklearn RandomForestRegressor.
    Predictions of the individual trees are interpreted as uncertainty.
    """

    def __init__(self, rf: RandomForestRegressor):
        """Constructs the model.

        Args:
            rf (RandomForestRegressor): Fitted sklearn random forest regressor.
        """
        super().__init__()
        if not isinstance(rf, RandomForestRegressor):
            raise ValueError("`rf` is not a sklearn RandomForestRegressor.")
        check_is_fitted(rf)
        self._rf = rf

    def forward(self, X: Tensor):
        r"""Compute the model output at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `batch_shape x s x n x m`-dimensional output tensor where
            `s` is the size of the ensemble.
        """
        # we transform to numpy
        nX = X.detach().numpy()
        # we need to check if we have a batch dimension
        if len(X.shape) != 3:
            # now we add the q-batch dimension
            nX = nX.reshape((1, *nX.shape))
        # loop over batches
        preds = []
        for i in range(nX.shape[0]):
            batch_preds = []
            # loop over estimators
            for estimator in self._rf.estimators_:
                batch_preds.append(
                    estimator.predict(nX[i]).reshape((nX[i].shape[0], 1))
                )
            preds.append(np.stack(batch_preds, axis=0))
        preds = np.stack(preds, axis=0)
        if X.ndim == 3:  # we have a batch dim
            return torch.from_numpy(preds).to(**tkwargs)
        else:  # we have no batch dim
            return torch.from_numpy(preds).to(**tkwargs).squeeze(dim=0)

    @property
    def num_outputs(self) -> int:
        r"""The number of outputs of the model."""
        return 1


class RandomForestSurrogate(BotorchSurrogate, TrainableSurrogate):
    """BoFire Random Forest model.

    The same hyperparameters are available as for the wrapped sklearn RandomForestRegreesor.
    """

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.n_estimators = data_model.n_estimators
        self.criterion = data_model.criterion
        self.max_depth = data_model.max_depth
        self.min_samples_split = data_model.min_samples_split
        self.min_samples_leaf = data_model.min_samples_leaf
        self.min_weight_fraction_leaf = data_model.min_weight_fraction_leaf
        self.max_features = data_model.max_features
        self.max_leaf_nodes = data_model.max_leaf_nodes
        self.min_impurity_decrease = data_model.min_impurity_decrease
        self.bootstrap = data_model.bootstrap
        self.oob_score = data_model.oob_score
        self.random_state = data_model.random_state
        self.ccp_alpha = data_model.ccp_alpha
        self.max_samples = data_model.max_samples
        super().__init__(data_model=data_model, **kwargs)

    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL
    model: Optional[_RandomForest] = None

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame):
        """Fit the Random Forest model.

        Args:
            X (pd.DataFrame): Dataframe with X values.
            Y (pd.DataFrame): Dataframe with Y values.
        """
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)
        rf = RandomForestRegressor(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,  # type: ignore
            min_samples_leaf=self.min_samples_leaf,  # type: ignore
            min_weight_fraction_leaf=self.min_weight_fraction_leaf,
            max_features=self.max_features,  # type: ignore
            max_leaf_nodes=self.max_leaf_nodes,
            min_impurity_decrease=self.min_impurity_decrease,
            bootstrap=self.bootstrap,
            oob_score=self.oob_score,
            random_state=self.random_state,
            ccp_alpha=self.ccp_alpha,
            max_samples=self.max_samples,
        )
        rf.fit(X=transformed_X.values, y=Y.values.ravel())
        self.model = _RandomForest(rf=rf)

    def _dumps(self) -> str:
        """Dumps the random forest to a string via pickle as this is not directly json serializable."""
        return codecs.encode(pickle.dumps(self.model._rf), "base64").decode()  # type: ignore

    def loads(self, data: str):
        """Loads the actual random forest from a base64 encoded pickle bytes object and writes it to the `model` attribute."""
        self.model = _RandomForest(
            rf=pickle.loads(codecs.decode(data.encode(), "base64"))
        )



import numpy as np
import pandas as pd

# Reaction Optimization Notebook util code

T0 = 25
T1 = 100
e0 = np.exp((T1 + 0) / T0)
e60 = np.exp((T1 + 60) / T0)
de = e60 - e0

boiling_points = {  # in °C
    "MeOH": 64.7,
    "THF": 66.0,
    "Dioxane": 101.0,
}
density = {  # in kg/l
    "MeOH": 0.792,
    "THF": 0.886,
    "Dioxane": 1.03,
}
# create dict from individual dicts
descs = {
    "boiling_points": boiling_points,
    "density": density,
}
solvent_descriptors = pd.DataFrame(descs)


# these functions are for faking real experimental data ;)
def calc_volume_fact(V):
    # 20-90
    # max at 75 = 1
    # min at 20 = 0.7
    # at 90=0.5
    x = (V - 20) / 70
    x = 0.5 + (x - 0.75) * 0.1 + (x - 0.4) ** 2
    return x


def calc_rhofact(solvent_type, Tfact):
    #  between 0.7 and 1.1
    x = solvent_descriptors["density"][solvent_type]
    x = (1.5 - x) * (Tfact + 0.5) / 2
    return x.values


def calc_Tfact(T):
    x = np.exp((T1 + T) / T0)
    return (x - e0) / de


# this can be used to create a dataframe of experiments including yields
def create_experiments(domain, nsamples=100, A=25, B=90, candidates=None):
    Tf = domain.inputs.get_by_key("Temperature")
    Vf = domain.inputs.get_by_key("Solvent Volume")
    typef = domain.inputs.get_by_key("Solvent Type")
    yf = domain.outputs.get_by_key("Yield")
    if candidates is None:
        T = np.random.uniform(low=Tf.lower_bound, high=Tf.upper_bound, size=(nsamples,))
        V = np.random.uniform(low=Vf.lower_bound, high=Vf.upper_bound, size=(nsamples,))
        solvent_types = [
            domain.inputs.get_by_key("Solvent Type").categories[np.random.randint(0, 3)]
            for i in range(nsamples)
        ]
    else:
        nsamples = len(candidates)
        T = candidates["Temperature"].values
        V = candidates["Solvent Volume"].values
        solvent_types = candidates["Solvent Type"].values

    Tfact = calc_Tfact(T)
    rhofact = calc_rhofact(solvent_types, Tfact)
    Vfact = calc_volume_fact(V)
    y = A * Tfact + B * rhofact
    y = 0.5 * y + 0.5 * y * Vfact
    # y = y.values
    samples = pd.DataFrame(
        {
            Tf.key: T,
            Vf.key: V,
            yf.key: y,
            typef.key: solvent_types,
            "valid_" + yf.key: np.ones(nsamples),
        },
        # index=pd.RangeIndex(nsamples),
    )
    samples.index = pd.RangeIndex(nsamples)
    return samples


def create_candidates(domain, nsamples=4):
    experiments = create_experiments(domain, nsamples=nsamples)
    candidates = experiments.drop(["Yield", "valid_Yield"], axis=1)
    return candidates


# this is for evaluating candidates that do not yet have a yield attributed to it.
def evaluate_experiments(domain, candidates):
    return create_experiments(domain, candidates=candidates)


evaluate_candidates = evaluate_experiments



from __future__ import annotations

from functools import total_ordering
from queue import PriorityQueue
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from bofire.data_models.constraints.api import ConstraintNotFulfilledError
from bofire.data_models.features.api import ContinuousInput
from bofire.strategies.doe.design import find_local_max_ipopt
from bofire.strategies.doe.objective import get_objective_class
from bofire.strategies.doe.utils import get_formula_from_string
from bofire.strategies.doe.utils_categorical_discrete import equal_count_split


@total_ordering
class NodeExperiment:
    def __init__(
        self,
        partially_fixed_experiments: pd.DataFrame,
        design_matrix: pd.DataFrame,
        value: float,
        categorical_groups: Optional[List[List[ContinuousInput]]] = None,
        discrete_vars: Optional[Dict[str, Tuple[ContinuousInput, List[float]]]] = None,
    ):
        """

        Args:
            partially_fixed_experiments: dataframe containing (some) fixed variables for experiments.
            design_matrix: optimal design for given the fixed and partially fixed experiments
            value: value of the objective function evaluated with the design_matrix
            categorical_groups: Represents the different groups of the categorical variables
            discrete_vars: Dict of discrete variables and the corresponding valid values in the optimization problem
        """
        self.partially_fixed_experiments = partially_fixed_experiments
        self.design_matrix = design_matrix
        self.value = value
        if categorical_groups is not None:
            self.categorical_groups = categorical_groups
        else:
            self.categorical_groups = []
        if discrete_vars is not None:
            self.discrete_vars = discrete_vars
        else:
            self.discrete_vars = {}

    def get_next_fixed_experiments(self) -> List[pd.DataFrame]:
        """
        Based on the current partially_fixed_experiment DataFrame the next branches are determined. One variable will
        be fixed more than before.
        Returns: List of the next possible branches where only one variable more is fixed

        """
        # branching for the binary/ categorical variables
        for group in self.categorical_groups:
            for row_index, _exp in self.partially_fixed_experiments.iterrows():
                if (
                    self.partially_fixed_experiments.iloc[row_index][group[0].key]
                    is None
                ):
                    current_keys = [elem.key for elem in group]
                    allowed_fixations = np.eye(len(group))
                    branches = [
                        self.partially_fixed_experiments.copy()
                        for i in range(len(allowed_fixations))
                    ]
                    for k, elem in enumerate(branches):
                        elem.loc[row_index, current_keys] = allowed_fixations[k]
                    return branches

        # branching for the discrete variables
        for key, (var, values) in self.discrete_vars.items():
            for row_index, _exp in self.partially_fixed_experiments.iterrows():
                current_fixation = self.partially_fixed_experiments.iloc[row_index][key]
                first_fixation, second_fixation = None, None
                if current_fixation is None:
                    lower_split, upper_split = equal_count_split(
                        values, var.lower_bound, var.upper_bound
                    )
                    first_fixation = (var.lower_bound, lower_split)
                    second_fixation = (upper_split, var.upper_bound)

                elif current_fixation[0] != current_fixation[1]:
                    lower_split, upper_split = equal_count_split(
                        values, current_fixation[0], current_fixation[1]
                    )
                    first_fixation = (current_fixation[0], lower_split)
                    second_fixation = (upper_split, current_fixation[1])

                if first_fixation is not None:
                    first_branch = self.partially_fixed_experiments.copy()
                    second_branch = self.partially_fixed_experiments.copy()

                    first_branch.loc[row_index, key] = first_fixation
                    second_branch.loc[row_index, key] = second_fixation

                    return [first_branch, second_branch]

        return []

    def __eq__(self, other: NodeExperiment) -> bool:
        return self.value == other.value

    def __ne__(self, other: NodeExperiment) -> bool:
        return self.value != other.value

    def __lt__(self, other: NodeExperiment) -> bool:
        return self.value < other.value

    def __str__(self):
        return (
            "\n ================ Branch-and-Bound Node ================ \n"
            + f"objective value: {self.value} \n"
            + f"design matrix: \n{self.design_matrix.round(4)} \n"
            + f"current fixations: \n{self.partially_fixed_experiments.round(4)} \n"
        )


def is_valid(node: NodeExperiment, tolerance: float = 1e-2) -> bool:
    """
    test if a design is a valid solution. i.e. binary and discrete variables are valid
    Args:
        node: the current node of the branch to be tested
        tolerance: absolute tolerance between valid values and values in the design

    Returns: True if the design is valid, else False

    """
    categorical_vars = [var for group in node.categorical_groups for var in group]
    design_matrix = node.design_matrix
    for var in categorical_vars:
        value = design_matrix.get(var.key)
        if not (
            np.logical_or(
                np.isclose(value, 0, atol=tolerance),
                np.isclose(value, 1, atol=tolerance),
            ).all()
        ):
            return False

    discrete_vars = node.discrete_vars
    for _key, (var, values) in discrete_vars.items():
        value = design_matrix.get(var.key)
        if False in [True in np.isclose(v, values, atol=tolerance) for v in value]:  # type: ignore
            return False
    return True


def bnb(
    priority_queue: PriorityQueue,
    verbose: bool = False,
    num_explored: int = 0,
    **kwargs,
) -> NodeExperiment:
    """
    branch-and-bound algorithm for solving optimization problems containing binary and discrete variables
    Args:
        num_explored: keeping track of how many branches have been explored
        priority_queue (PriorityQueue): initial nodes of the branching tree
        verbose (bool): if true, print information during the optimization process
        **kwargs: parameters for the actual optimization / find_local_max_ipopt

    Returns: a branching Node containing the best design found

    """
    if priority_queue.empty():
        raise RuntimeError("Queue empty before feasible solution was found")

    domain = kwargs["domain"]
    n_experiments = kwargs["n_experiments"]

    # get objective function
    model_formula = get_formula_from_string(
        model_type=kwargs["model_type"], rhs_only=True, domain=domain
    )
    objective_class = get_objective_class(kwargs["objective"])
    objective_class = objective_class(
        domain=domain, model=model_formula, n_experiments=n_experiments
    )

    pre_size = priority_queue.qsize()
    current_branch = priority_queue.get()
    # test if current solution is already valid
    if is_valid(current_branch):
        return current_branch

    # branch current solutions in sub-problems
    next_branches = current_branch.get_next_fixed_experiments()

    if verbose:
        print(
            f"current length of branching queue (+ new branches): {pre_size} + {len(next_branches)} currently "
            f"explored branches: {num_explored}, current best value: {current_branch.value}"
        )
    # solve branched problems
    for _i, branch in enumerate(next_branches):
        kwargs["sampling"] = current_branch.design_matrix
        try:
            design = find_local_max_ipopt(partially_fixed_experiments=branch, **kwargs)
            value = objective_class.evaluate(design.to_numpy().flatten())
            new_node = NodeExperiment(
                branch,
                design,
                value,
                current_branch.categorical_groups,
                current_branch.discrete_vars,
            )
            domain.validate_candidates(
                candidates=design.apply(lambda x: np.round(x, 8)),
                only_inputs=True,
                tol=1e-4,
                raise_validation_error=True,
            )

            priority_queue.put(new_node)
        except ConstraintNotFulfilledError:
            if verbose:
                print("skipping branch because of not fulfilling constraints")

    return bnb(
        priority_queue,
        verbose=verbose,
        num_explored=num_explored + len(next_branches),
        **kwargs,
    )
